from unittest import TestCase
from unittest.mock import Mock

from musicscore import Chord, Beat, Part
from musicscore.exceptions import ChordTypeNotSetError, ChordTestError


class TestTestChordNumberOfBeams(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.chord = Chord(60, 1 / 8)

    def test_chord_has_no_type(self):
        with self.assertRaises(ChordTypeNotSetError):
            self.chord.check_number_of_beams()

    def test_chord_without_beam(self):
        self.chord.type = "quarter"
        assert self.chord.check_number_of_beams()
        self.chord.type = "32nd"
        assert self.chord.check_number_of_beams()
        self.chord.beams = {}
        assert self.chord.check_number_of_beams()
        self.chord.beams = {1: "begin"}
        with self.assertRaises(ChordTestError):
            self.chord.check_number_of_beams()
        self.chord.beams = {1: "begin", 3: "forward"}
        with self.assertRaises(ChordTestError):
            self.chord.check_number_of_beams()
        self.chord.beams = {1: "begin", 2: "begin", 3: "forward"}
        assert self.chord.check_number_of_beams()
        self.chord.beams = {1: "begin", 2: "begin", 3: "begin", 4: "forward"}
        with self.assertRaises(ChordTestError):
            self.chord.check_number_of_beams()


class TestBeams16th(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.beat = Beat()
        self.measure = Mock()
        self.beat._parent = self.measure
        [self.beat._add_chord(Chord(60, qd)) for qd in 4 * [1 / 4]]
        self.chords = self.beat.get_chords()

    def test_all_non_rest(self):
        assert [ch.beams for ch in self.chords] == [{}, {}, {}, {}]
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {1: "begin", 2: "begin"},
            {1: "continue", 2: "continue"},
            {1: "continue", 2: "continue"},
            {1: "end", 2: "end"},
        ]

    def test_all_two_rests_middle(self):
        for ch in self.chords[1:3]:
            ch.midis = 0
        assert [ch.is_rest for ch in self.chords] == [False, True, True, False]
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {1: "begin", 2: "forward"},
            {},
            {},
            {1: "end", 2: "backward"},
        ]

    def test_one_rest_second(self):
        self.chords[1].midis = 0
        assert [ch.is_rest for ch in self.chords] == [False, True, False, False]
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {1: "begin", 2: "forward"},
            {},
            {1: "continue", 2: "begin"},
            {1: "end", 2: "end"},
        ]

    def test_one_rest_third(self):
        self.chords[2].midis = 0
        assert [ch.is_rest for ch in self.chords] == [False, False, True, False]
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {1: "begin", 2: "begin"},
            {1: "continue", 2: "end"},
            {},
            {1: "end", 2: "backward"},
        ]

    def test_one_rest_start(self):
        self.chords[0].midis = 0
        assert [ch.is_rest for ch in self.chords] == [True, False, False, False]
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {},
            {1: "begin", 2: "begin"},
            {1: "continue", 2: "continue"},
            {1: "end", 2: "end"},
        ]

    def test_two_rests_start_end(self):
        self.chords[0].midis = 0
        self.chords[-1].midis = 0
        assert [ch.is_rest for ch in self.chords] == [True, False, False, True]
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {},
            {1: "begin", 2: "begin"},
            {1: "end", 2: "end"},
            {},
        ]

    def test_all_rests(self):
        for ch in self.chords:
            ch.midis = 0
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [{}, {}, {}, {}]

    def test_break_all_beams(self):
        for ch in self.chords:
            ch.break_beam()
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [{}, {}, {}, {}]

    def test_break_last_beam(self):
        self.chords[-1].break_beam()
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {1: "begin", 2: "begin"},
            {1: "continue", 2: "continue"},
            {1: "end", 2: "end"},
            {},
        ]

    def test_break_third_beam(self):
        self.chords[2].break_beam()
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {1: "begin", 2: "begin"},
            {1: "end", 2: "end"},
            {1: "begin", 2: "begin"},
            {1: "end", 2: "end"},
        ]

    def test_break_third_and_fourth_beam(self):
        self.chords[2].break_beam()
        self.chords[3].break_beam()
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {1: "begin", 2: "begin"},
            {1: "end", 2: "end"},
            {},
            {},
        ]

    def test_manually_set_beams_to_None(self):
        for ch in self.chords:
            ch.beams = None
        self.beat._update_chord_beams()
        # assert [ch.beams for ch in self.chords] == [{}, {}, {}, {}]
        ch = self.chords[0]
        ch.get_parent_measure().get_divisions.return_value = 4
        ch.up.up.up.number = 1
        self.chords[0].finalize()


class TestBeams32th(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.beat = Beat()
        self.beat._parent = Mock()
        [self.beat._add_chord(Chord(60, qd)) for qd in 8 * [1 / 8]]
        self.chords = self.beat.get_chords()

    def test_second_and_third(self):
        self.chords[1].midis = 0
        self.chords[2].midis = 0
        self.beat._update_chord_beams()
        assert [ch.beams for ch in self.chords] == [
            {1: "begin", 2: "forward", 3: "forward"},
            {},
            {},
            {1: "continue", 2: "backward", 3: "backward"},
            {1: "continue", 2: "begin", 3: "begin"},
            {1: "continue", 2: "continue", 3: "continue"},
            {1: "continue", 2: "continue", 3: "continue"},
            {1: "end", 2: "end", 3: "end"},
        ]

    def test_32nd_two_chords(self):
        p = Part("p1")
        p.add_measure([1, 8])
        qds = [1 / 8, 3 / 8]
        [p.add_chord(Chord(71, qd)) for qd in qds]
        p.finalize()
        chords = p.get_chords()
        assert [ch.beams for ch in chords] == [
            {1: "begin", 2: "begin", 3: "forward"},
            {1: "end", 2: "end"},
        ]

    def test_32nd_problem_1(self):
        p = Part("p2")
        p.add_measure([1, 4])
        qds = [1 / 8, 3 / 8, 1 / 2]
        [p.add_chord(Chord(71, qd)) for qd in qds]
        p.finalize()
        chords = p.get_chords()
        for ch in chords:
            ch.check_number_of_beams()
        assert [ch.beams for ch in chords] == [
            {1: "begin", 2: "begin", 3: "forward"},
            {1: "continue", 2: "end"},
            {1: "end"},
        ]

    def test_32nd_problem_2(self):
        p = Part("p1")
        p.add_measure([1, 8])
        qds = [3 / 8, 1 / 8]
        [p.add_chord(Chord(71, qd)) for qd in qds]
        p.finalize()
        chords = p.get_chords()
        assert [ch.beams for ch in chords] == [
            {1: "begin", 2: "begin"},
            {1: "end", 2: "end", 3: "backward"},
        ]

    def test_32nd_problem_3(self):
        p = Part("p2")
        p.add_measure([1, 4])
        qds = [3 / 8, 1 / 8, 1 / 2]
        [p.add_chord(Chord(71, qd)) for qd in qds]
        p.finalize()
        chords = p.get_chords()
        for ch in chords:
            ch.check_number_of_beams()
        assert [ch.beams for ch in chords] == [
            {1: "begin", 2: "begin"},
            {1: "continue", 2: "end", 3: "backward"},
            {1: "end"},
        ]

    def test_32nd_problem_4(self):
        p = Part("p2")
        p.add_measure([1, 4])
        qds = [1 / 2, 1 / 8, 3 / 8]
        [p.add_chord(Chord(71, qd)) for qd in qds]
        p.finalize()
        chords = p.get_chords()
        for ch in chords:
            ch.check_number_of_beams()
        assert [ch.beams for ch in chords] == [
            {1: "begin"},
            {1: "continue", 2: "begin", 3: "forward"},
            {1: "end", 2: "end"},
        ]



"""
A note can be the end of one slur and the start of a new slur. Also, in MusicXML, nested slurs are possible like
in the second measure where one slur goes over all four notes, and another slur goes from the second to the third
note."""

from pathlib import Path

from musicscore import Score, A, C, G, Chord
from unittest import TestCase
from musicscore.util import slur_chords


class TestLily33c(TestCase):
    def test_lily_33c_Spanners_Slurs(self):
        score = Score()
        part = score.add_part("p1")
        midis = [G(4), C(5), A(4), G(4), G(4), C(5), A(4), G(4)]
        chords = [Chord(m, 1) for m in midis]
        slur_chords(chords[:2])
        slur_chords(chords[1:3])
        slur_chords(chords[2:4])

        slur_chords(chords[4:])
        slur_chords(chords[5:7], number=2)
        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



class MusicTreeException(Exception):
    pass


class MusicTreeTypeError(MusicTreeException, TypeError):
    pass


class ClassHasNoMusicXMLEquivalentError(MusicTreeException, AttributeError):
    pass


class AddChordError(MusicTreeException):
    def __init__(self):
        msg = "Use Part.add_chord() instead!"
        super().__init__(msg)


class AlreadyFinalizedError(MusicTreeException):
    def __init__(self, object_, method_=None):
        msg = f"{object_.__class__.__name__} is already finalized."
        if method_:
            msg += f" Method {object_.__class__.__name__}.{method_}() cannot be called after finalization."
        super().__init__(msg)


class DeepCopyException(MusicTreeException):
    pass


class WrongNumberOfChordsError(MusicTreeException):
    pass


# Beat exceptions


class BeatException(MusicTreeException):
    pass


class BeatWrongDurationError(BeatException, ValueError):
    pass


class BeatIsFullError(BeatException):
    pass


class BeatNotFullError(BeatException):
    pass


class BeatHasNoParentError(BeatException):
    pass


class BeatUpdateChordTupletsError(BeatException):
    pass


# Chord exceptions
class ChordException(MusicTreeException):
    pass


class ChordAddXException(ChordException):
    pass


class ChordAddXPlacementException(ChordAddXException):
    pass


class ChordAlreadySplitError(ChordException):
    pass


class ChordAlreadyHasNotesError(ChordException):
    pass


class ChordCannotSplitError(ChordException):
    pass


class ChordHasNoMidisError(ChordException):
    pass


class ChordHasNoParentBeamError(ChordException):
    pass


class ChordHasNoParentPartError(ChordException):
    pass


class ChordHasNoNotesError(ChordException):
    pass


class ChordHasNoQuarterDurationError(ChordException):
    pass


class ChordNumberOfDotsNotSetError(ChordException):
    pass


class ChordParentBeamError(ChordException):
    pass


class ChordQuarterDurationAlreadySetError(ChordException):
    pass


class ChordTestError(ChordException):
    pass


class ChordTypeNotSetError(ChordException):
    pass


# GraceChord Exceptions


class GraceChordException(ChordException):
    pass


class GraceChordCannotHaveGraceNotesError(GraceChordException):
    pass


class GraceChordCannotSetQuarterDurationError(GraceChordException):
    pass


# Lyrics exceptions


class LyricsException(MusicTreeException):
    pass


class LyricsExtensionError(LyricsException):
    pass


class LyricSyllabicOrExtensionError(LyricsException):
    pass


class LyricsWrongNumberOfChordsError(LyricsException):
    pass


# Measure exceptions
class MeasureException(MusicTreeException):
    pass


# Metronome exceptions
class MetronomeException(MusicTreeException):
    pass


class MetronomeWrongBeatUnitError(MetronomeException):
    pass


# Note exceptions
class NoteException(MusicTreeException):
    pass


class NotationException(NoteException):
    pass


class NoteHasNoParentChordError(NoteException):
    pass


class NoteMidiHasNoParentChordError(NoteException):
    pass


class NoteTypeError(NoteException):
    pass


# Part exceptions
class PartException(MusicTreeException):
    pass


# QuarterDuration exceptions
class QuarterDurationException(MusicTreeException):
    pass


class QuarterDurationIsNotWritable(QuarterDurationException):
    pass


# Rest exceptions


class RestException(MusicTreeException):
    pass


class RestCannotSetMidiError(RestException):
    pass


class RestWithDisplayStepHasNoDisplayOctave(RestException):
    pass


class RestWithDisplayOctaveHasNoDisplayStep(RestException):
    pass


# SimpleFormat exceptions
class SimpleFormatException(MusicTreeException):
    pass


# Score exceptions
class ScoreException(MusicTreeException):
    pass


class ScoreMultiMeasureRestError(ScoreException):
    pass


class ScorePartIdIsNotUniqueError(ScoreException):
    pass


class ScoreHasNoPartsError(ScoreException):
    pass


# Staff exceptions
class StaffException(MusicTreeException):
    pass


class StaffHasNoParentError(StaffException):
    pass


# Time exceptions
class TimeException(MusicTreeException):
    pass


class TimeActualSignaturesNotValidError(TimeException):
    pass


# Tuplet exceptions


class TupletException(MusicTreeException):
    pass


class TupletNormalTypeError(TupletException):
    pass


# Voice exceptions
class VoiceException(MusicTreeException):
    pass


class VoiceHasNoBeatsError(VoiceException):
    pass


class VoiceHasNoParentError(VoiceException):
    pass


class VoiceIsFullError(VoiceException):
    pass



from pathlib import Path
from unittest import TestCase

from musicscore import Chord
from musicscore.exceptions import WrongNumberOfChordsError
from musicscore.midi import C

from musicscore.score import Score
from musicscore.tests.util import diff_xml, create_test_objects
from musicscore.util import (
    lcm,
    isinstance_as_string,
    XML_DYNAMIC_CLASSES,
    XML_ARTICULATION_CLASSES,
    XML_ORNAMENT_CLASSES,
    XML_ORNAMENT_AND_OTHER_NOTATIONS,
    XML_TECHNICAL_CLASSES,
    XML_DIRECTION_TYPE_CLASSES,
    XML_OTHER_NOTATIONS,
    XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS,
    slur_chords,
    split_list,
    wedge_chords,
    trill_chords,
    bracket_chords,
    octave_chords,
)
from musicxml import XMLTrillMark, XMLWavyLine, XMLOctaveShift


class TestUtils(TestCase):
    def test_isinstance_as_string(self):
        assert isinstance_as_string(Score(), "Score")
        assert isinstance_as_string(Score(), "MusicTree")
        assert not isinstance_as_string(Score(), "str")

        class Measure:
            pass

        assert not isinstance_as_string(Measure(), "MusicTree")

        assert isinstance_as_string(C(4), "MidiNote")
        assert isinstance_as_string(C(4), "Midi")
        assert isinstance_as_string(C(4), "C")
        assert not isinstance_as_string(C(4), "str")

    def test_diff_xml_no_diff(self):
        path = Path(__file__).parent / "test_util_diff_xml.xml"
        assert diff_xml(path) == []

    def test_diff_xml_with_diff(self):
        path = Path(__file__).parent / "test_util_diff_xml.xml"
        path2 = Path(__file__).parent / "test_util_diff_xml_with_diff.xml"
        assert diff_xml(path, path2) == ["- <key>", "- <fifths>0</fifths>", "- </key>"]

    def test_lcm(self):
        assert lcm([3, 4, 5, 7]) == 420
        assert lcm([2, 4, 6]) == 12

    def test_slur_chords(self):
        chords = [Chord(60, 1)]
        with self.assertRaises(WrongNumberOfChordsError):
            slur_chords(chords)
        chords.extend([Chord(61, 1), Chord(62, 1)])
        slur_chords(chords)
        assert chords[0].get_slurs()[0].type == "start"
        assert chords[1].get_slurs()[0].type == "continue"
        assert chords[2].get_slurs()[0].type == "stop"

    def test_wedge_chords(self):
        chords = [Chord(60, 1)]
        with self.assertRaises(WrongNumberOfChordsError):
            wedge_chords(chords, "crescendo")
        chords.extend([Chord(61, 1), Chord(62, 1)])
        wedge_chords(chords, "crescendo")
        assert chords[0].get_wedges()[0].type == "crescendo"
        assert chords[1].get_wedges()[0].type == "continue"
        assert chords[2].get_wedges()[0].type == "stop"

    def test_trill_chords(self):
        chords = [Chord(60, 1)]
        with self.assertRaises(WrongNumberOfChordsError):
            trill_chords(chords)
        chords.extend([Chord(61, 1), Chord(62, 1)])
        trill_chords(chords)
        assert len(chords[0].get_x(XMLTrillMark)) == 1
        assert len(chords[1].get_x(XMLTrillMark)) == 0
        assert len(chords[2].get_x(XMLTrillMark)) == 0
        assert chords[0].get_x(XMLWavyLine)[0].type == "start"
        assert chords[1].get_x(XMLWavyLine)[0].type == "continue"
        assert chords[2].get_x(XMLWavyLine)[0].type == "stop"

    def test_bracket_chords(self):
        chords = [Chord(60, 1)]
        with self.assertRaises(WrongNumberOfChordsError):
            bracket_chords(chords)
        chords.extend([Chord(61, 1), Chord(62, 1)])
        bracket_chords(chords)
        assert chords[0].get_brackets()[0].type == "start"
        assert chords[1].get_brackets()[0].type == "continue"
        assert chords[2].get_brackets()[0].type == "stop"

        for i, ch in enumerate(chords):
            bracket = ch.get_brackets()[0]
            assert bracket.line_type == "solid"
            # assert bracket.placement == 'above'
            assert bracket.number == 1
            if i == 0:
                assert bracket.line_end == "down"
            elif i == 1:
                assert bracket.line_end == "none"
            else:
                assert bracket.line_end == "down"

        chords = [Chord(60, 1) for _ in range(3)]
        bracket_chords(chords, "dashed", "none", "up", placement="below", number=2)
        for i, ch in enumerate(chords):
            bracket = ch.get_brackets()[0]
            # assert bracket.placement == 'below'
            assert bracket.number == 2
            assert bracket.line_type == "dashed"
            if i == 0:
                assert bracket.line_end == "none"
            elif i == 1:
                assert bracket.line_end == "none"
            else:
                assert bracket.line_end == "up"

    def test_octave_chords(self):
        chords = [Chord(60, 1) for _ in range(3)]
        octave_chords(chords)
        assert chords[0].get_x(XMLOctaveShift)[0].type == "down"
        assert chords[1].get_x(XMLOctaveShift)[0].type == "continue"
        assert chords[2].get_x(XMLOctaveShift)[0].type == "stop"

        chords = [Chord(60, 1) for _ in range(3)]
        octave_chords(chords, type="down", size=15, number=2)
        assert chords[0].get_x(XMLOctaveShift)[0].type == "down"
        assert chords[1].get_x(XMLOctaveShift)[0].type == "continue"
        assert chords[2].get_x(XMLOctaveShift)[0].type == "stop"
        for ch in chords:
            assert ch.get_x(XMLOctaveShift)[0].size == 15
            assert ch.get_x(XMLOctaveShift)[0].number == 2

    def test_split_list(self):
        original_list = [1, 2, 3, 4, 5, 6]
        split_indices = [2, 5]
        self.assertListEqual(
            split_list(original_list, split_indices), [[1, 2], [3, 4, 5], [6]]
        )


class TestTestObjects(TestCase):
    def test_direction_type_test_objects(self):
        test_object = create_test_objects(type="direction_type")
        assert len(
            XML_DIRECTION_TYPE_CLASSES + XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS
        ) == len(test_object)
        for obj in test_object:
            assert (
                obj.__class__ in XML_DIRECTION_TYPE_CLASSES
                or obj.__class__ in XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS
            )

    def test_ornament_test_objects(self):
        test_object = create_test_objects(type="ornament")
        assert len(XML_ORNAMENT_CLASSES + XML_ORNAMENT_AND_OTHER_NOTATIONS) == len(
            test_object
        )
        for obj in test_object:
            assert (
                obj.__class__ in XML_ORNAMENT_CLASSES + XML_ORNAMENT_AND_OTHER_NOTATIONS
            )

    def test_technical_test_objects(self):
        test_object = create_test_objects(type="technical")
        assert len(XML_TECHNICAL_CLASSES) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_TECHNICAL_CLASSES

    def test_articulation_test_objects(self):
        test_object = create_test_objects(type="articulation")
        assert len(XML_ARTICULATION_CLASSES) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_ARTICULATION_CLASSES

    def test_other_notation_test_objects(self):
        test_object = create_test_objects(type="notation")
        assert len(XML_OTHER_NOTATIONS + XML_ORNAMENT_AND_OTHER_NOTATIONS) == len(
            test_object
        )
        for obj in test_object:
            assert (
                obj.__class__ in XML_OTHER_NOTATIONS + XML_ORNAMENT_AND_OTHER_NOTATIONS
            )

    def test_dynamics_test_objects(self):
        test_object = create_test_objects(type="dynamics")
        assert len(XML_DYNAMIC_CLASSES) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_DYNAMIC_CLASSES



"""
Some grace notes and after-graces
"""

from pathlib import Path

from musicscore import Score, Chord, A, E, G
from unittest import TestCase


class TestLily24d(TestCase):
    def test_lily_24d_AfterGrace(self):
        score = Score()
        part = score.add_part("p1")

        chords = [Chord(E(5), 2), Chord(E(5), 2)]
        for midi in [G(5), A(5), A(5)]:
            chords[0].add_grace_chord(midi, type="16th", position="after")

        chords[1].add_grace_chord(G(5), type="16th", position="after")
        chords[1].add_grace_chord(A(5), type="16th", position="after")

        for ch in chords:
            part.add_chord(ch)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Multiple (simple) lyrics. The order of the exported stanzas is relevant (identified by the
number attribute in this test case)
"""

from pathlib import Path

from musicscore import Score, G, Chord, Lyrics
from unittest import TestCase


class TestLily61b(TestCase):
    def test_lily_61b_Lyrics(self):
        score = Score()
        part = score.add_part("p1")
        chords = [Chord(G(4), 1) for _ in range(8)]
        lyrics_1 = [
            (
                "Tra",
                "la",
                "la,",
            ),
            ("ja!", None),
            ("Tra", "ra"),
            "...",
        ]
        lyrics_2 = [
            (
                "Tra",
                "la",
                "la,",
            ),
            ("ja!", None),
            ("Tra", "ra"),
            "...",
        ]
        lyrics_3 = [
            (
                "TRA",
                "LA",
                "LA,",
            ),
            ("JA!", None),
            ("TRA", "Ra"),
            "...",
        ]
        Lyrics(lyrics_1, number=1, show_number=True).add_to_chords(chords)
        Lyrics(lyrics_2, number=2, show_number=True).add_to_chords(chords)
        Lyrics(lyrics_3, number=3, show_number=True).add_to_chords(chords)
        [part.add_chord(ch) for ch in chords]
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Two properly nested part groups: One group (with a square bracket) goes from staff 2 to 4)
and another group (with a curly bracket) goes from staff 3 to 4.
"""

from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase


class TestLily41d(TestCase):
    def test_lily_41d_StaffGroups(self):
        score = Score()
        parts = [score.add_part(f"p-{i}") for i in range(1, 6)]
        for p in parts:
            p.add_chord(Chord(0, 4))

        score.group_parts(1, 2, 4, symbol="square")
        score.group_parts(2, 3, 4, symbol="bracket")

        # new_xml_part_list = XMLPartList(xsd_check=False)
        #
        # for index, xml_score_part in enumerate(score.xml_part_list.get_children()):
        #     if index == 1:
        #         pg = new_xml_part_list.add_child(XMLPartGroup(number='1', type='start'))
        #         pg.add_child(XMLGroupSymbol('square'))
        #         pg.add_child(XMLGroupBarline('yes'))
        #     elif index == 2:
        #         pg = new_xml_part_list.add_child(XMLPartGroup(number='2', type='start'))
        #         pg.add_child(XMLGroupSymbol('bracket'))
        #         pg.add_child(XMLGroupBarline('yes'))
        #
        #     new_xml_part_list.add_child(xml_score_part)
        #
        #     if index == 3:
        #         new_xml_part_list.add_child(XMLPartGroup(number='1', type='stop'))
        #         new_xml_part_list.add_child(XMLPartGroup(number='2', type='stop'))
        # score.xml_part_list = new_xml_part_list
        path = Path(__file__).with_suffix(".xml")
        score.export_xml(path)



from typing import Optional, Union

from musicxml import XMLAccidental

from musicscore.musictree import MusicTree
from musicscore.xmlwrapper import XMLWrapper

__all__ = [
    "STANDARD",
    "FLAT",
    "SHARP",
    "ENHARMONIC",
    "FORCESHARP",
    "FORCEFLAT",
    "SIGNS",
    "Accidental",
]
#:
STANDARD = {
    0: ("C", 0, 0),
    0.5: ("C", 0.5, 0),
    1: ("C", 1, 0),
    1.5: ("C", 1.5, 0),
    2: ("D", 0, 0),
    2.5: ("E", -1.5, 0),
    3: ("E", -1, 0),
    3.5: ("E", -0.5, 0),
    4: ("E", 0, 0),
    4.5: ("E", 0.5, 0),
    5: ("F", 0, 0),
    5.5: ("F", 0.5, 0),
    6: ("F", 1, 0),
    6.5: ("F", 1.5, 0),
    7: ("G", 0, 0),
    7.5: ("A", -1.5, 0),
    8: ("A", -1, 0),
    8.5: ("A", -0.5, 0),
    9: ("A", 0, 0),
    9.5: ("B", -1.5, 0),
    10: ("B", -1, 0),
    10.5: ("B", -0.5, 0),
    11: ("B", 0, 0),
    11.5: ("C", -0.5, 1),
}

#:
FLAT = {
    0: ("C", 0, 0),
    0.5: ("D", -1.5, 0),
    1: ("D", -1, 0),
    1.5: ("D", -0.5, 0),
    2: ("D", 0, 0),
    2.5: ("E", -1.5, 0),
    3: ("E", -1, 0),
    3.5: ("E", -0.5, 0),
    4: ("E", 0, 0),
    4.5: ("F", -0.5, 0),
    5: ("F", 0, 0),
    5.5: ("G", -1.5, 0),
    6: ("G", -1, 0),
    6.5: ("G", -0.5, 0),
    7: ("G", 0, 0),
    7.5: ("A", -1.5, 0),
    8: ("A", -1, 0),
    8.5: ("A", -0.5, 0),
    9: ("A", 0, 0),
    9.5: ("B", -1.5, 0),
    10: ("B", -1, 0),
    10.5: ("B", -0.5, 0),
    11: ("B", 0, 0),
    11.5: ("C", -0.5, 1),
}

#:
SHARP = {
    0: ("C", 0, 0),
    0.5: ("C", 0.5, 0),
    1: ("C", 1, 0),
    1.5: ("C", 1.5, 0),
    2: ("D", 0, 0),
    2.5: ("D", 0.5, 0),
    3: ("D", 1, 0),
    3.5: ("D", 1.5, 0),
    4: ("E", 0, 0),
    4.5: ("E", 0.5, 0),
    5: ("F", 0, 0),
    5.5: ("F", 0.5, 0),
    6: ("F", 1, 0),
    6.5: ("F", 1.5, 0),
    7: ("G", 0, 0),
    7.5: ("G", 0.5, 0),
    8: ("G", 1, 0),
    8.5: ("G", 1.5, 0),
    9: ("A", 0, 0),
    9.5: ("A", 0.5, 0),
    10: ("A", 1, 0),
    10.5: ("A", 1.5, 0),
    11: ("B", 0, 0),
    11.5: ("B", 0.5, 0),
}

#:
ENHARMONIC = {
    0: ("C", 0, 0),
    0.5: ("D", -1.5, 0),
    1: ("D", -1, 0),
    1.5: ("D", -0.5, 0),
    2: ("D", 0, 0),
    2.5: ("D", 0.5, 0),
    3: ("D", 1, 0),
    3.5: ("D", 1.5, 0),
    4: ("E", 0, 0),
    4.5: ("F", -0.5, 0),
    5: ("F", 0, 0),
    5.5: ("G", -1.5, 0),
    6: ("G", -1, 0),
    6.5: ("G", -0.5, 0),
    7: ("G", 0, 0),
    7.5: ("G", 0.5, 0),
    8: ("G", 1, 0),
    8.5: ("G", 1.5, 0),
    9: ("A", 0, 0),
    9.5: ("A", 0.5, 0),
    10: ("A", 1, 0),
    10.5: ("A", 1.5, 0),
    11: ("B", 0, 0),
    11.5: ("B", 0.5, 0),
}

#:
FORCESHARP = {
    0: ("B", 1, -1),
    0.5: ("B", 1.5, -1),
    1: ("B", 2, -1),
    1.5: ("C", 1.5, 0),
    2: ("C", 2, 0),
    2.5: ("D", 0.5, 0),
    3: ("D", 1, 0),
    3.5: ("D", 1.5, 0),
    4: ("D", 2, 0),
    4.5: ("E", 0.5, 0),
    5: ("E", 1, 0),
    5.5: ("E", 1.5, 0),
    6: ("E", 2, 0),
    6.5: ("F", 1.5, 0),
    7: ("F", 2, 0),
    7.5: ("G", 0.5, 0),
    8: ("G", 1, 0),
    8.5: ("G", 1.5, 0),
    9: ("G", 2, 0),
    9.5: ("A", 0.5, 0),
    10: ("A", 1, 0),
    10.5: ("A", 1.5, 0),
    11: ("A", 2, 0),
    11.5: ("B", 0.5, 0),
}

#:
FORCEFLAT = {
    0: ("D", -2, 0),
    0.5: ("D", -1.5, 0),
    1: ("D", -1, 0),
    1.5: ("D", -0.5, 0),
    2: ("E", -2, 0),
    2.5: ("E", -1.5, 0),
    3: ("F", -2, 0),
    3.5: ("F", -1.5, 0),
    4: ("F", -1, 0),
    4.5: ("F", -0.5, 0),
    5: ("G", -2, 0),
    5.5: ("G", -1.5, 0),
    6: ("G", -1, 0),
    6.5: ("G", -0.5, 0),
    7: ("A", -2, 0),
    7.5: ("A", -1.5, 0),
    8: ("A", -1, 0),
    8.5: ("A", -0.5, 0),
    9: ("B", -2, 0),
    9.5: ("B", -1.5, 0),
    10: ("C", -2, 1),
    10.5: ("C", -1.5, 1),
    11: ("C", -1, 1),
    11.5: ("C", -0.5, 1),
}

#:
SIGNS = {
    -2: "flat-flat",
    -1.5: "three-quarters-flat",
    -1: "flat",
    -0.5: "quarter-flat",
    0: "natural",
    0.5: "quarter-sharp",
    1: "sharp",
    1.5: "three-quarters-sharp",
    2: "double-sharp",
}


class Accidental(MusicTree, XMLWrapper):
    """
    Parent type: :obj:`~musicscore.midi.Midi`

    Child type: None

    Accidental is the class for managing :obj:`musicscore.midi.Midi`'s accidental sign and its pitch parameters: step, alter, octave.
    The parameter mode ('standard', 'enharmonic', 'flat', 'sharp', 'force-flat', 'force-sharp') can be used to set different enharmonic variants of the same
    pitch.
    """

    _ATTRIBUTES = {"mode", "show", "parent_midi"}

    XMLClass = XMLAccidental

    def __init__(self, mode="standard", show: Optional[bool] = None, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(value_="natural", **kwargs)
        self._mode = None
        self._show = None
        self.show = show
        self.mode = mode

    def _update(self):
        self._update_parent_midi()
        self._update_xml_object()

    def _update_parent_midi(self):
        if self.parent_midi and self.parent_midi.value != 0:
            self.parent_midi._update_pitch_parameters()

    def _update_xml_object(self):
        if self.sign:
            self._xml_object.value_ = self.sign

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> Optional[XMLClass]:
        """
        If an attribute is not found directly in a :obj:`~musicscore.musictree.MusicTree` class which inherits this class, it wll be passed on to ``__get_attribute__`` and ``__set__attribute__`` methods of ``xml_object``.
        It can also use the short cut attributes of :obj:`~musicxml.xmlelement.xmlelement.XMLElement` to get or set the first child.

        :return: wrapped MusicXML element of type :obj:`XMLClass`
        """
        if self.parent_midi and self.parent_midi.value == 0:
            return None
        if self.show is True:
            return self._xml_object
        elif self.show is False:
            return None

    @property
    def mode(self):
        """
        permitted modes: ``standard``, ``flat``, ``sharp``, ``enharmonic``, ``force-sharp``, ``force-flat``

        :return: accidental mode which corresponds to global variables :obj:`~musicscore.accidental.STANDARD`, :obj:`~musicscore.accidental.FLAT`, :obj:`~musicscore.accidental.SHARP`,
                 :obj:`~musicscore.accidental.ENHARMONIC`, :obj:`~musicscore.accidental.FORCESHARP`, :obj:`~musicscore.accidental.FORCEFLAT`
        """
        return self._mode

    @mode.setter
    def mode(self, value):
        permitted = (
            "standard",
            "flat",
            "sharp",
            "enharmonic",
            "force-sharp",
            "force-flat",
        )
        if value not in permitted:
            raise TypeError(f"accidental_mode.value {value} must be in {permitted}")
        self._mode = value
        self._update_xml_object()
        self._update_parent_midi()

    @property
    def parent_midi(self) -> "Midi":
        """
        :return: The midi parent of Accidental in the :obj:`~musicscore.musictree.MusicTree` structure. It is equivalent to :obj:`up` or :obj:`get_parent()`
        :rtype: :obj:`musicscore.midi.Midi`
        """
        return self.up

    @parent_midi.setter
    def parent_midi(self, val):
        val.add_child(self)

    @property
    def sign(self) -> Optional[str]:
        """
        Converts ``alter`` parameter of :obj:`~get_pitch_parameters` into an actual ``sign`` depending on :obj:`~mode`. :obj:`~parent_midi` must be set first.

        :return: Possible values: ``flat-flat``, ``three-quarters-flat``, ``flat``, ``quarter-flat``, ``natural``,
                 ``quarter-sharp``, ``sharp``, ``three-quarters-sharp``, ``double-sharp``
        """
        try:
            alter = self.get_pitch_parameters()[1]
            return SIGNS[alter]
        except TypeError:
            return None

    @property
    def show(self) -> bool:
        """
        If ``False`` ``xml_object`` will be ``None`` and no accidental is shown in the :obj:`~musicscore.score.Score`.
        If ``True`` the accidental is shown. If the value of one of the following xml_object :obj:`~musicxml.xmlelement.xmlelement.XMLAccidental` properties:``self.parentheses``, ``self.editorial``.
        ``self.cautionary`` or ``self.bracket`` is set to yes, the accidental is always shown regardless of this
        property
        """
        if "yes" in [self.parentheses, self.editorial, self.cautionary, self.bracket]:
            return True
        return self._show

    @show.setter
    def show(self, val):
        if val is not None and not isinstance(val, bool):
            raise TypeError
        if val != self._show:
            self._show = val
            try:
                self.up.up._update_xml_accidental()
            except AttributeError:
                pass

    def get_pitch_parameters(
        self, midi_value: Optional[Union[int, float]] = None
    ) -> Optional[tuple]:
        """
        :param: a valid midi value or ``None``. If ``midi_value == 0`` (for a rest) return value is ``None``. If ``midi_value is None`` and a :obj:`~parent_midi` exists, parent_midi's value will be used.

        :return: A tuple consisting of pitch step name, alter value and octave value: ``(step, alter, octave)``

        .. seealso::
           :obj:`~mode`
        """

        if midi_value is None:
            if self.parent_midi:
                midi_value = self.parent_midi.value
            else:
                return None
        if midi_value == 0:
            return None

        if self.mode == "standard":
            output = STANDARD[midi_value % 12]

        elif self.mode == "enharmonic":
            output = ENHARMONIC[midi_value % 12]

        elif self.mode == "force-sharp":
            output = FORCESHARP[midi_value % 12]

        elif self.mode == "force-flat":
            output = FORCEFLAT[midi_value % 12]

        elif self.mode == "flat":
            output = FLAT[midi_value % 12]

        elif self.mode == "sharp":
            output = SHARP[midi_value % 12]

        else:
            raise ValueError
        return output[0], output[1], output[2] + (int(midi_value // 12)) - 1

    def __copy__(self):
        return self.__class__(mode=self.mode, show=self.show)



"""How to treat lyrics and slurred notes. Normally, a slurred group of notes is assigned only one lyrics syllable."""

from pathlib import Path

from musicscore import Score, A, Chord, Lyrics, C, E
from unittest import TestCase
from musicscore.util import slur_chords


class TestLily61d(TestCase):
    def test_lily_61d_Lyrics_Melisma(self):
        score = Score()
        part = score.add_part("p1")

        midis = [C(5), (A(4), E(5)), C(5), (C(5), E(5)), C(5), C(5), C(5), E(5)]
        chords = [Chord(m, 1) for m in midis]

        [part.add_chord(ch) for ch in chords]

        slur_chords(chords[:4])
        slur_chords(chords[4:6])
        slur_chords(chords[6:])

        lyrics = [("Me", None, None, None, "lis", None, "ma.", None)]
        Lyrics(lyrics).add_to_chords(chords)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



import inspect
from unittest import skip

from musicscore import Score
from unittest import TestCase

from musicscore.tests.util import generate_path, generate_repetitions


class TestCautionaryAccidentalSigns(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_part("p.")

    @skip
    def test_cautionary_signs_repetition_traditional(self):
        self.part.show_accidental_signs = "traditional"
        generate_repetitions(self.part)
        path = generate_path(inspect.currentframe())
        self.score.export_xml(path)

    @skip
    def test_cautionary_signs_repetition_modern(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()

    @skip
    def test_cautionary_signs_traditional(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()

    @skip
    def test_cautionary_signs_modern(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()

    @skip
    def test_cautionary_signs_different_octaves_traditional(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()

    def test_cautionary_signs_different_octaves_modern(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()

    def test_cautionary_signs_different_voice_traditional(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()

    def test_cautionary_signs_different_voice_modern(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()

    def test_cautionary_signs_multi_staff_traditional(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()

    def test_cautionary_signs_multi_staff_modern(self):
        path = generate_path(inspect.currentframe())
        print(path)
        # self.fail()



"""
A clef change in the middle of a measure, using either an implicit measure or simply placing
the attributes in the middle of the measure.
"""

from pathlib import Path

from musicscore import Score, Chord, Clef, TrebleClef
from unittest import TestCase


class TestLily46c(TestCase):
    def test_lily_46c_Midmeasure_Clef(self):
        score = Score()
        part = score.add_part("p1")
        part.add_chord(Chord(0, 4))
        [part.add_chord(Chord(72, 1)) for _ in range(8)]
        ch = part.get_measure(2).get_chords()[2]
        ch.clef = Clef(sign="C", line=2)
        ch = part.get_measure(3).get_chords()[2]
        ch.clef = TrebleClef()

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from unittest import TestCase
from unittest.mock import patch, Mock

from fractions import Fraction

from musicscore.chord import Chord
from musicscore.exceptions import NoteMidiHasNoParentChordError
from musicscore.midi import Midi
from musicscore.note import Note, tie, untie
from musicscore.part import Part
from musicxml.xmlelement.xmlelement import *


class NoteTestCase(TestCase):
    def setUp(self) -> None:
        self.mock_chord = Mock(spec=Chord)
        self.mock_chord.get_voice_number.return_value = 1
        self.mock_chord.get_staff_number.return_value = None
        self.mock_chord.number_of_dots = 0
        self.mock_chord.tuplet = None
        self.mock_chord.beams = {}
        self.mock_measure = Mock()
        self.mock_measure.get_divisions.return_value = 1
        self.mock_chord.get_parent_measure.return_value = self.mock_measure
        self.mock_chord.type = "quarter"

    def tearDown(self):
        patch.stopall()

    def set_parent_chord(self, midi):
        if not isinstance(midi, Midi):
            midi = Midi(midi)
        midi._set_parent_chord(self.mock_chord)
        return midi


class TestNote(NoteTestCase):
    def test_mock_chord(self):
        assert self.mock_chord.get_voice_number() == 1
        assert self.mock_chord.get_staff_number() is None
        assert self.mock_chord.get_parent_measure().get_divisions() == 1

    def test_note_init_with_midi(self):
        with self.assertRaises(TypeError):
            Note()

        with self.assertRaises(NoteMidiHasNoParentChordError):
            Note(midi=Midi(60))

    def test_note_init(self):
        n = Note(midi=self.set_parent_chord(71))
        self.mock_chord.type = "half"
        n.quarter_duration = 2
        n.midi.accidental.show = False
        expected = """<note>
  <pitch>
    <step>B</step>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
</note>
"""
        assert n.to_string() == expected
        self.mock_chord.get_voice_number.return_value = 2
        n = Note(midi=self.set_parent_chord(61), default_x=10)
        self.mock_measure.get_divisions.return_value = 2
        n.quarter_duration = 2
        n.xml_notehead = XMLNotehead("square")
        assert n.midi.value == 61
        n.midi.accidental.show = True
        expected = """<note default-x="10">
  <pitch>
    <step>C</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
  <duration>4</duration>
  <voice>2</voice>
  <type>half</type>
  <accidental>sharp</accidental>
  <notehead>square</notehead>
</note>
"""
        assert n.to_string() == expected

    def test_note_set_divisions(self):
        self.mock_measure.get_divisions.return_value = 3
        n = Note(midi=self.set_parent_chord(61), quarter_duration=Fraction(1, 3))
        assert n.xml_duration.value_ == 1
        self.mock_measure.get_divisions.return_value = 6
        n._update_xml_duration()
        assert n.xml_duration.value_ == 2

    def test_note_type(self):
        self.mock_chord.type = "half"
        n = Note(midi=self.set_parent_chord(60), quarter_duration=2)
        n.midi.accidental.show = False
        expected = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
</note>
"""
        assert n.to_string() == expected

    def test_note_dots(self):
        self.mock_measure.get_divisions.return_value = 2
        self.mock_chord.number_of_dots = 1
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1.5)
        n._update_xml_dots()
        n.midi.accidental.show = False
        expected = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>3</duration>
  <voice>1</voice>
  <type>quarter</type>
  <dot />
</note>
"""
        assert n.to_string() == expected
        assert len(n.xml_object.find_children("XMLDot")) == 1
        self.mock_chord.number_of_dots = 0
        n._update_xml_dots()
        assert len(n.xml_object.find_children("XMLDot")) == 0
        self.mock_chord.number_of_dots = 3
        n._update_xml_dots()
        assert len(n.xml_object.find_children("XMLDot")) == 3

    def test_change_midi_or_duration(self):
        self.mock_chord.get_voice_number.return_value = 2
        self.mock_chord.type = "half"
        n = Note(midi=self.set_parent_chord(61), quarter_duration=2, default_x=10)
        n.midi.accidental.show = True
        n.xml_notehead = "square"
        n.midi.accidental.show = True
        expected = """<note default-x="10">
  <pitch>
    <step>C</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>2</voice>
  <type>half</type>
  <accidental>sharp</accidental>
  <notehead>square</notehead>
</note>
"""
        assert n.to_string() == expected
        n.midi.value = 62
        n.midi.accidental.show = False
        expected = """<note default-x="10">
  <pitch>
    <step>D</step>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>2</voice>
  <type>half</type>
  <notehead>square</notehead>
</note>
"""
        assert n.midi.up == n
        assert n.to_string() == expected
        n.midi.value = 0
        expected = """<note default-x="10">
  <rest />
  <duration>2</duration>
  <voice>2</voice>
  <type>half</type>
</note>
"""
        assert n.to_string() == expected
        n.midi.value = 72
        n.xml_notehead = XMLNotehead("diamond")
        expected = """<note default-x="10">
  <pitch>
    <step>C</step>
    <octave>5</octave>
  </pitch>
  <duration>2</duration>
  <voice>2</voice>
  <type>half</type>
  <notehead>diamond</notehead>
</note>
"""
        assert n.to_string() == expected
        n.midi.value = 0
        expected = """<note default-x="10">
  <rest />
  <duration>2</duration>
  <voice>2</voice>
  <type>half</type>
</note>
"""
        assert n.to_string() == expected
        # n.midi = None
        # with self.assertRaises(XMLElementChildrenRequired):
        #     n.to_string()
        n.midi = self.set_parent_chord(Midi(80))
        n.midi.accidental.show = True
        n.quarter_duration = 0
        expected = """<note default-x="10">
  <grace />
  <pitch>
    <step>A</step>
    <alter>-1</alter>
    <octave>5</octave>
  </pitch>
  <voice>2</voice>
  <type>half</type>
  <accidental>flat</accidental>
</note>
"""
        assert n.to_string() == expected
        with self.assertRaises(ValueError):
            n.midi = self.set_parent_chord(Midi(0))

    def test_grace_note(self):
        self.mock_chord.type = None
        n = Note(midi=self.set_parent_chord(61), quarter_duration=0, default_x=10)
        n.midi.accidental.show = True
        n.xml_notehead = XMLNotehead("square")
        expected = """<note default-x="10">
  <grace />
  <pitch>
    <step>C</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
  <voice>1</voice>
  <accidental>sharp</accidental>
  <notehead>square</notehead>
</note>
"""
        assert n.xml_object.to_string() == expected

    def test_cue_note(self):
        self.mock_chord.type = "half"
        n = Note(midi=self.set_parent_chord(Midi(61)), quarter_duration=2)
        n.midi.accidental.show = True
        n.xml_cue = XMLCue()
        expected = """<note>
  <cue />
  <pitch>
    <step>C</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
  <accidental>sharp</accidental>
</note>
"""
        assert n.xml_object.to_string() == expected

    def test_note_stem(self):
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n.midi.accidental.show = False
        n.xml_stem = "up"
        expected = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <voice>1</voice>
  <type>quarter</type>
  <stem>up</stem>
</note>
"""
        assert n.to_string() == expected

    def test_note_hide_show_accidental(self):
        n = Note(midi=self.set_parent_chord(60))
        n.midi.accidental.show = True
        assert n.xml_accidental.value_ == "natural"
        n.midi.accidental.show = False
        assert n.midi.accidental.xml_object is None
        assert n.xml_object.xml_accidental is None

    def test_note_up_chord(self):
        self.mock_chord.get_voice_number.return_value = 1
        n = Note(midi=self.set_parent_chord(61), quarter_duration=1)
        assert n.up == self.mock_chord

    def test_note_midi_change_accidental_show(self):
        self.mock_chord.get_voice_number.return_value = 1
        n = Note(midi=self.set_parent_chord(61), quarter_duration=2, default_x=10)
        n.midi.accidental.show = True
        assert n.xml_object.xml_accidental == n.midi.accidental.xml_object
        assert (
            n.xml_object.xml_accidental.value_
            == n.midi.accidental.xml_object.value_
            == "sharp"
        )
        n.midi.value = 62
        assert (
            n.xml_object.xml_accidental.value_
            == n.midi.accidental.xml_object.value_
            == "natural"
        )
        n.midi.accidental.show = False
        assert n.xml_object.xml_accidental == n.midi.accidental.xml_object

    def test_note_update_xml_notations(self):
        n = Note(midi=self.set_parent_chord(60))
        n.get_or_create_xml_notations()
        n.xml_notations.xml_articulations = XMLArticulations()
        accent = n.xml_notations.xml_articulations.add_child(XMLAccent())
        staccato = n.xml_notations.xml_articulations.add_child(XMLStaccato())
        n.xml_notations.xml_technical = XMLTechnical()
        string = n.xml_notations.xml_technical.add_child(XMLString(1))
        assert len(n.xml_notations.get_children()) == 2
        assert len(n.xml_notations.xml_articulations.get_children()) == 2
        assert len(n.xml_notations.xml_technical.get_children()) == 1

        n.xml_notations.xml_articulations.remove(accent)
        assert len(n.xml_notations.get_children()) == 2
        assert len(n.xml_notations.xml_articulations.get_children()) == 1
        assert len(n.xml_notations.xml_technical.get_children()) == 1

        n.xml_notations.xml_articulations.remove(staccato)

        assert len(n.xml_notations.get_children()) == 2
        assert len(n.xml_notations.xml_articulations.get_children()) == 0
        assert len(n.xml_notations.xml_technical.get_children()) == 1

        n._update_xml_notations()
        assert len(n.xml_notations.get_children()) == 1
        assert n.xml_notations.xml_articulations is None
        assert len(n.xml_notations.xml_technical.get_children()) == 1

        n.xml_notations.xml_technical.remove(string)
        assert len(n.xml_notations.get_children()) == 1
        assert n.xml_notations.xml_articulations is None
        assert len(n.xml_notations.xml_technical.get_children()) == 0

        n._update_xml_notations()
        assert n.xml_notations is None


standard_note_xml = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <voice>1</voice>
  <type>quarter</type>
</note>
"""
standard_note_xml_start_tie = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <tie type="start" />
  <voice>1</voice>
  <type>quarter</type>
  <notations>
    <tied type="start" />
  </notations>
</note>
"""

standard_note_xml_stop_tie = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <tie type="stop" />
  <voice>1</voice>
  <type>quarter</type>
  <notations>
    <tied type="stop" />
  </notations>
</note>
"""
standard_note_xml_stop_start_tie = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <tie type="stop" />
  <tie type="start" />
  <voice>1</voice>
  <type>quarter</type>
  <notations>
    <tied type="stop" />
    <tied type="start" />
  </notations>
</note>
"""


class TestNoteTie(NoteTestCase):
    def test_tie_manually(self):
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n.xml_object.add_child(XMLTie(type="start"))
        n.xml_notations = XMLNotations()
        n.xml_notations.add_child(XMLTied(type="start"))
        assert n.is_tied
        assert not n.is_tied_to_previous
        assert n.to_string() == standard_note_xml_start_tie

    def test_remove_tie(self):
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n.remove_tie()
        n.start_tie()
        n.remove_tie("start")
        n.stop_tie()
        n.remove_tie("stop")

        n.midi.add_tie("start")
        assert n.is_tied

    def test_start_tie(self):
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n.start_tie()
        assert n.is_tied
        assert n.midi._ties == {"start"}
        assert not n.is_tied_to_previous
        assert n.to_string() == standard_note_xml_start_tie
        n.remove_tie()
        assert not n.is_tied
        assert n.midi._ties == set()
        assert not n.is_tied_to_previous
        assert n.to_string() == standard_note_xml

    def test_stop_tie(self):
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n.stop_tie()
        assert n.midi._ties == {"stop"}
        assert not n.is_tied
        assert n.is_tied_to_previous
        assert n.to_string() == standard_note_xml_stop_tie
        n.remove_tie()
        assert n.midi._ties == set()
        assert not n.is_tied
        assert not n.is_tied_to_previous
        assert n.to_string() == standard_note_xml

    def test_start_stop_tie(self):
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n.start_tie()
        n.stop_tie()
        assert n.midi._ties == {"start", "stop"}
        assert n.is_tied
        assert n.is_tied_to_previous
        assert n.to_string() == standard_note_xml_stop_start_tie

        n.remove_tie("start")
        assert n.midi._ties == {"stop"}
        assert not n.is_tied
        assert n.is_tied_to_previous
        assert n.to_string() == standard_note_xml_stop_tie
        n.remove_tie()
        assert n.midi._ties == set()
        assert not n.is_tied
        assert not n.is_tied_to_previous
        assert n.to_string() == standard_note_xml

    # def test_tie_accidentals(self):
    #     n = Note(midi=61, quarter_duration=1)
    #     n.stop_tie()
    #     n.start_tie()
    #     assert n.midi.accidental.show is False
    #     assert n.xml_object.xml_accidental is None
    #
    #     n.remove_tie('start')
    #     assert n.midi.accidental.show is False
    #     n.remove_tie()
    #     assert n.midi.accidental.show is None

    def test_tie_untie_one_note(self):
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        tie(n)
        assert n.to_string() == standard_note_xml_start_tie
        untie(n)
        assert n.to_string() == standard_note_xml

    def test_tie_untie_two_notes(self):
        n1 = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n2 = Note(midi=self.set_parent_chord(60), quarter_duration=1)

        tie(n1, n2)
        assert (
            n1.to_string() + n2.to_string()
            == standard_note_xml_start_tie + standard_note_xml_stop_tie
        )
        untie(n1, n2)
        assert n1.to_string() + n2.to_string() == standard_note_xml + standard_note_xml

    def test_tie_untie_tree_or_more_notes(self):
        n1 = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n2 = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n3 = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n4 = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        tie(n1, n2, n3, n4)
        assert (
            n1.to_string() + n2.to_string() + n3.to_string() + n4.to_string()
            == standard_note_xml_start_tie
            + standard_note_xml_stop_start_tie
            + standard_note_xml_stop_start_tie
            + standard_note_xml_stop_tie
        )
        untie(n1, n2, n3, n4)
        assert (
            n1.to_string() + n2.to_string() + n3.to_string() + n4.to_string()
            == standard_note_xml
            + standard_note_xml
            + standard_note_xml
            + standard_note_xml
        )

    def test_note_staff_number(self):
        n1 = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        assert n1.get_staff_number() is None
        assert n1.xml_object.xml_staff is None
        self.mock_chord.get_staff_number.return_value = 1
        assert n1.get_staff_number() == 1
        n1._update_xml_staff()
        assert n1.xml_object.xml_staff.value_ == 1

    def test_note_parent_chord(self):
        """
        Test if parent chord of a note is its parent in the musicscore
        """
        p = Part("p1")
        ch = Chord(60, 1)
        p.add_chord(ch)
        ch.finalize()
        n = ch.notes[0]
        assert n.up == n.parent_chord

    def test_note_update_ties(self):
        n = Note(midi=self.set_parent_chord(60), quarter_duration=1)
        n.midi.add_tie("start")
        assert n.is_tied_to_next
        assert not n.is_tied_to_previous
        n.midi.add_tie("stop")
        assert n.is_tied_to_previous
        n.midi.remove_tie("start")
        n.remove_tie("start")
        assert not n.is_tied_to_next
        n.midi.remove_tie("stop")
        assert not n.is_tied_to_previous



from pathlib import Path

from musicscore import Score, C, Chord, Time
from unittest import TestCase

"""Accidentals can be cautionary or editorial. Each measure has a normal accidental, an editorial, a cautionary and 
an editorial and cautionary accidental."""


class TestLily01e(TestCase):
    def test_lily_01e_Pitches_ParenthesizedAccidentals(self):
        score = Score("cautionary accidentals")
        p = score.add_part("cautionary")
        p.add_measure(Time(5, 4))
        midis = (
            [C(4, "bb") for _ in range(5)]
            + [C(4, "three-quarters-flat") for _ in range(5)]
            + [C(4, "b") for _ in range(5)]
            + [C(4, "quarter-flat") for _ in range(5)]
            + [C(4, "natural") for _ in range(5)]
            + [C(4, "quarter-sharp") for _ in range(5)]
            + [C(4, "#") for _ in range(5)]
            + [C(4, "three-quarters-sharp") for _ in range(5)]
            + [C(4, "##") for _ in range(5)]
        )
        for index, midi in enumerate(midis):
            if (index + 1) % 5 == 0:
                midi.accidental.bracket = "yes"
                ch = Chord(midi, 1)
                ch.add_lyric("bracket")
            elif (index + 1) % 5 == 1:
                ch = Chord(midi, 1)
            elif (index + 1) % 5 == 2:
                midi.accidental.parentheses = "yes"
                ch = Chord(midi, 1)
                ch.add_lyric("parentheses")
            elif (index + 1) % 5 == 3:
                midi.accidental.cautionary = "yes"
                ch = Chord(midi, 1)
                ch.add_lyric("cautionary")
            else:
                midi.accidental.editorial = "yes"
                ch = Chord(midi, 1)
                ch.add_lyric("editorial")
            p.add_chord(ch)

        xml_file = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_file)



from unittest import TestCase
from musicscore import Part, Chord, Measure, Beat, Voice, Score, Staff, Accidental
from musicscore.exceptions import AlreadyFinalizedError, AddChordError
from musicxml.xmlelement.xmlelement import XMLCoda, XMLSegno


class TestAddExceptions(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_part("p1")
        self.measure = self.part.add_measure()
        self.staff = self.measure.add_staff()
        self.voice = self.staff.add_voice()
        self.beat = self.voice.add_beat()

    def test_add_chord_to_part(self):
        p = Part(id="part-1")
        p.add_chord(Chord(60, 4))
        p.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            p.add_chord(Chord(60, 4))

    def test_add_child_to_part(self):
        p = Part(id="part-1")
        p.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            p.add_child(Measure(1))

    def test_add_measure_to_part(self):
        p = Part(id="part-1")
        p.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            p.add_measure()

    def test_add_direction_type_to_chord(self):
        p = Part(id="part-1")
        ch = Chord(60, 4)
        p.add_chord(ch)
        ch.add_direction_type(XMLCoda())
        ch.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            ch.add_direction_type(XMLSegno())

    def test_add_midi_to_chord(self):
        p = Part(id="part-1")
        ch = Chord(60, 4)
        p.add_chord(ch)
        ch.add_midi(63)
        assert [m.value for m in ch.midis] == [60, 63]
        ch.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            ch.add_midi(65)

    def test_add_dynamics_to_chord(self):
        p = Part(id="part-1")
        ch = Chord(60, 4)
        p.add_chord(ch)
        ch.add_dynamics("ff")
        ch.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            ch.add_dynamics("pp")

    def test_add_tie_to_chord(self):
        ch = Chord([60, 62], 4)
        self.part.add_chord(ch)
        ch.add_tie("start")
        ch.finalize()
        for note in ch.notes:
            assert note.is_tied_to_next
            assert not note.is_tied_to_previous
        ch.add_tie("stop")
        for note in ch.notes:
            assert note.is_tied_to_next
            assert note.is_tied_to_previous

    def test_add_lyric_to_chord(self):
        p = Part(id="part-1")
        ch = Chord(60, 1)
        p.add_chord(ch)
        lyrics1 = ch.add_lyric("one")
        lyrics2 = ch.add_lyric("two")
        ch.finalize()
        assert ch.notes[0].find_children("XMLLyric") == [lyrics1, lyrics2]

    def test_add_wedge_to_chord(self):
        p = Part(id="part-1")
        ch = Chord(60, 4)
        p.add_chord(ch)
        ch.add_wedge("crescendo")
        ch.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            ch.add_wedge("stop")

    def test_add_child_to_measure(self):
        self.measure.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.measure.add_child(Staff())

    def test_add_staff_to_measure(self):
        self.measure.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.measure.add_staff()

    def test_add_voice_to_measure(self):
        self.measure.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.measure.add_voice()

    def test_add_child_to_beat(self):
        self.beat.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.beat.add_child(Chord(60, 1))

    def test_add_chord_to_beat(self):
        self.beat.finalize()
        with self.assertRaises(AddChordError):
            self.beat.add_chord(Chord(60, 4))

    def test_add_child_to_midi(self):
        ch = Chord(60, 1)
        self.part.add_chord(ch)
        ch.midis[0].add_child(Accidental())
        ch.number_of_dots = 1
        ch.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            ch.midis[0].add_child(Accidental())

    def test_add_tie_to_midi(self):
        ch = Chord([60, 62], 4)
        self.part.add_chord(ch)
        ch.midis[0].add_tie("start")
        ch.finalize()
        n1, n2 = ch.notes
        assert n1.is_tied_to_next
        assert not n1.is_tied_to_previous
        assert not n2.is_tied_to_next
        assert not n2.is_tied_to_previous
        ch.midis[0].add_tie("stop")
        assert n1.is_tied_to_next
        assert n1.is_tied_to_previous
        assert not n2.is_tied_to_next
        assert not n2.is_tied_to_previous

    def test_add_child_to_score(self):
        self.score.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.score.add_child(Part("p2"))

    def test_add_part_to_score(self):
        self.score.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.score.add_part("p2")

    def test_add_child_to_staff(self):
        self.staff.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.staff.add_child(Voice())

    def test_add_voice_to_staff(self):
        self.staff.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.staff.add_voice()

    def test_add_child_to_voice(self):
        self.voice.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.voice.add_child(Beat(1))

    def test_add_chord_to_voice(self):
        self.voice.finalize()
        with self.assertRaises(AddChordError):
            self.voice.add_chord(Chord(60, 1))

    def test_add_beat_to_voice(self):
        self.voice.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.voice.add_beat(1)



from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase

"""
A simple, repeated measure (repeated 5 times)
"""


class TestLily45a(TestCase):
    def test_lily_45a_SimpleRepeat(self):
        score = Score()
        part = score.add_part("p1")

        [part.add_chord(Chord(0, 4)) for _ in range(2)]
        part.get_measure(1).set_repeat_barline(times=5)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from typing import Optional

from musicscore.exceptions import TupletNormalTypeError
from musicxml import XMLTimeModification, XMLTuplet

TUPLETACTUALTONORMALNOTES = {
    3: 2,
    5: 4,
    6: 4,
    7: 4,
    9: 8,
    10: 8,
    11: 8,
    12: 8,
    13: 8,
    14: 8,
    15: 8,
}
TUPLETNORMALTYPES = {8: "32nd", 4: "16th", 2: "eighth", 1: "quarter", 0.5: "half"}


class Tuplet:
    def __init__(
        self,
        actual_notes=3,
        normal_notes=None,
        normal_type=None,
        bracket_type=None,
        bracket_number=None,
        quarter_duration=1,
    ):
        self._actual_notes = None
        self._normal_notes = None
        self._normal_type = None
        self._bracket_type = None
        self._bracket_number = None
        self._quarter_duration = None
        self.actual_notes = actual_notes
        self.normal_notes = normal_notes
        self.normal_type = normal_type
        self.bracket_type = bracket_type
        self.bracket_number = bracket_number
        self.quarter_duration = quarter_duration

    @property
    def actual_notes(self) -> int:
        return self._actual_notes

    @actual_notes.setter
    def actual_notes(self, val):
        if not isinstance(val, int):
            raise TypeError
        if val < 2:
            raise ValueError
        self._actual_notes = val

    @property
    def normal_notes(self) -> int:
        if self._normal_notes is None:
            return TUPLETACTUALTONORMALNOTES.get(self.actual_notes)
        return self._normal_notes

    @normal_notes.setter
    def normal_notes(self, val):
        if val is not None:
            if not isinstance(val, int):
                raise TypeError
            if val < 2:
                raise ValueError
        self._normal_notes = val

    @property
    def normal_type(self) -> str:
        if self._normal_type is None:
            try:
                return TUPLETNORMALTYPES[self.normal_notes / self.quarter_duration]
            except KeyError:
                raise TupletNormalTypeError(
                    f"Cannot find normal type of normal_notes {self.normal_notes} with quarter_duration {self.quarter_duration}. You can set normal_type manually."
                )
        return self._normal_type

    @normal_type.setter
    def normal_type(self, val):
        permitted = [
            None,
            "1024th",
            "512th",
            "256th",
            "128th",
            "64th",
            "32nd",
            "16th",
            "eighth",
            "quarter",
            "half",
            "whole",
            "breve",
            "long",
            "maxima",
        ]
        if val not in permitted:
            raise ValueError(f"Permitted type values are: {permitted}")
        self._normal_type = val

    @property
    def bracket_type(self) -> Optional[str]:
        return self._bracket_type

    @bracket_type.setter
    def bracket_type(self, val):
        permitted = [None, "start", "stop"]
        if val not in permitted:
            raise ValueError(f"Permitted bracket_type values are: {permitted}")
        self._bracket_type = val

    @property
    def bracket_number(self) -> int:
        if self._bracket_number is None and self.bracket_type:
            return 1
        return self._bracket_number

    @bracket_number.setter
    def bracket_number(self, val):
        if val is not None:
            if not isinstance(val, int):
                raise TypeError
            if val < 1:
                raise ValueError
        self._bracket_number = val

    @property
    def ratio(self):
        return (self.actual_notes, self.normal_notes)

    def get_xml_time_modification(self) -> "XMLTimeModification":
        output = XMLTimeModification()
        output.xml_actual_notes = self.actual_notes
        output.xml_normal_notes = self.normal_notes
        output.xml_normal_type = self.normal_type
        return output

    def get_xml_tuplet(self) -> XMLTuplet:
        if self.bracket_type:
            output = XMLTuplet(type=self.bracket_type, number=self.bracket_number)
            if self.bracket_type == "start":
                output.bracket = "yes"
            return output



from pathlib import Path

from unittest import TestCase
from musicxml.xmlelement.xmlelement import (
    XMLRehearsal,
    XMLDynamics,
    XMLOtherDynamics,
    XMLSegno,
    XMLCoda,
    XMLWords,
    XMLEyeglasses,
    XMLDashes,
    XMLBracket,
    XMLOctaveShift,
    XMLPedal,
    XMLMetronome,
    XMLBeatUnit,
    XMLPerMinute,
    XMLHarpPedals,
    XMLPedalStep,
    XMLPedalAlter,
    XMLPedalTuning,
    XMLDamp,
    XMLDampAll,
    XMLScordatura,
    XMLAccord,
    XMLTuningStep,
    XMLTuningOctave,
    XMLTuningAlter,
    XMLAccordionRegistration,
    XMLAccordionHigh,
    XMLAccordionMiddle,
    XMLAccordionLow,
    XMLWedge,
    XMLPrint,
    XMLBarline,
    XMLBarStyle,
)
from musicscore import Score, Chord

"""
All <direction> elements defined in MusicXML.
"""

""""
XML_DIRECTION_TYPE_CLASSES = [
    XMLSymbol, XMLStringMute, XMLImage, XMLPrincipalVoice, XMLPercussion, XMLStaffDivide, XMLOtherDirection
]
"""


class TestLily31a(TestCase):
    def test_lily_31a_Directions(self):
        score = Score(title="MusicXML directions (attached to staff)")
        part = score.add_part("p1")

        for rehearsal, enclosure in zip(
            ["A", "B", "Test", "Crc"], ["square", "none", "rectangle", "circle"]
        ):
            r = XMLRehearsal(rehearsal, enclosure=enclosure)
            ch = Chord(60, 1)
            ch.add_direction_type(r)
            part.add_chord(ch)

        for dt in [XMLSegno(), XMLCoda(), XMLWords("words"), XMLEyeglasses()]:
            ch = Chord(60, 1)
            if isinstance(dt, XMLWords):
                ch.add_direction_type(dt, "below")
            else:
                ch.add_direction_type(dt, "above")
            part.add_chord(ch)

        dynamics = [
            "p",
            "pp",
            "ppp",
            "pppp",
            "ppppp",
            "pppppp",
            "f",
            "ff",
            "fff",
            "ffff",
            "fffff",
            "ffffff",
            "mp",
            "mf",
            "sf",
            "sfp",
            "sfpp",
            "fp",
            "rf",
            "rfz",
            "sfz",
            "sffz",
            "fz",
        ]
        for d in dynamics:
            ch = Chord(60, 1)
            ch.add_dynamics(d)
            part.add_chord(ch)

        ch = Chord(60, 1)
        d = XMLDynamics()
        d.add_child(XMLOtherDynamics("abc-ffz"))
        ch.add_direction_type(d, placement="below")
        part.add_chord(ch)

        for w in ["crescendo", "stop"]:
            ch = Chord(60, 1)
            ch.add_wedge(w)
            part.add_chord(ch)

        for x in ["start", "stop"]:
            ch = Chord(60, 1)
            ch.add_direction_type(XMLDashes(type=x))
            part.add_chord(ch)

        for x in ["start", "stop"]:
            ch = Chord(60, 1)
            ch.add_direction_type(XMLBracket(type=x, line_end="down"))
            part.add_chord(ch)

        for x in ["down", "stop"]:
            ch = Chord(60, 1)
            ch.add_direction_type(XMLOctaveShift(type=x))
            part.add_chord(ch)

        for x in ["start", ("stop", "start"), "continue", "stop"]:
            ch = Chord(60, 1)
            if isinstance(x, tuple):
                ch.add_direction_type(
                    XMLPedal(type=x[0], relative_x=-25, relative_y=-10), "below"
                )
                ch.add_direction_type(XMLPedal(type=x[1], relative_y=-10), "below")
            else:
                ch.add_direction_type(XMLPedal(type=x, relative_y=-10), "below")
            part.add_chord(ch)

        ch = Chord(60, 1)
        m = XMLMetronome()
        m.add_child(XMLBeatUnit("quarter"))
        m.add_child(XMLPerMinute("60"))
        ch.add_direction_type(m)
        part.add_chord(ch)

        ch = Chord(60, 1)
        steps = ["D", "C", "B", "E", "F", "G", "A"]
        alters = [0, -1, -1, 0, 0, 1, -1]
        hp = XMLHarpPedals(relative_y=-20, font_size=14)
        for s, a in zip(steps, alters):
            pt = hp.add_child(XMLPedalTuning())
            pt.add_child(XMLPedalStep(s))
            pt.add_child(XMLPedalAlter(a))
        ch.add_direction_type(hp, "below")
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_direction_type(XMLDamp())
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_direction_type(XMLDampAll())
        part.add_chord(ch)

        """
        <scordatura>
           <accord string="1">
              <tuning-step>E</tuning-step>
              <tuning-octave>5</tuning-octave>
           </accord>
           <accord string="2">
              <tuning-step>A</tuning-step>
              <tuning-octave>4</tuning-octave>
           </accord>
           <accord string="3">
              <tuning-step>E</tuning-step>
              <tuning-alter>-1</tuning-alter>
              <tuning-octave>4</tuning-octave>
           </accord>
           <accord string="4">
              <tuning-step>A</tuning-step>
              <tuning-octave>3</tuning-octave>
           </accord>
        </scordatura>
        """

        sc = XMLScordatura()
        steps = ["E", "A", "E", "A"]
        alters = [None, None, -1, None]
        octaves = [5, 4, 4, 3]
        for index, (s, a, o) in enumerate(zip(steps, alters, octaves)):
            st = sc.add_child(XMLAccord(string=index + 1))
            st.add_child(XMLTuningStep(s))
            if a:
                st.add_child(XMLTuningAlter(a))
            st.add_child(XMLTuningOctave(o))

        ch = Chord(60, 1)
        ch.add_direction_type(sc)
        ch.add_lyric("Scordatura")
        part.add_chord(ch)

        """
        <direction>
           <direction-type>
              <accordion-registration>
                 <accordion-high/>
                 <accordion-middle>3</accordion-middle>
              </accordion-registration>
           </direction-type>
        </direction>
        """
        ch = Chord(60, 1)
        ag = XMLAccordionRegistration()
        ag.add_child(XMLAccordionHigh())
        ag.add_child(XMLAccordionMiddle(2))
        ag.add_child(XMLAccordionLow())
        ch.add_direction_type(ag)
        part.add_chord(ch)

        ch = Chord(0, 2)
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_dynamics("p")
        ch.add_direction_type(
            XMLWords("subito", font_style="italic", relative_y=-20), placement="below"
        )
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_dynamics("ppp")
        ch.add_wedge(XMLWedge(type="crescendo", relative_x=25))
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_dynamics("fff")
        ch.add_wedge(XMLWedge(type="stop", relative_x=-5))
        part.add_chord(ch)

        ch = Chord(0, 1)
        part.add_chord(ch)

        new_systems = [2, 5, 9, 12]
        for m in new_systems:
            part.get_measure(m).xml_object.add_child(XMLPrint(new_system="yes"))

        barline = XMLBarline(location="right")
        barline.add_child(XMLBarStyle("light-light"))
        part.get_measure(13).xml_object.add_child(barline)

        barline = XMLBarline(location="right")
        barline.add_child(XMLBarStyle("light-heavy"))
        part.get_measure(14).xml_object.add_child(barline)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
All different types of glissando defined in MusicXML
"""

from pathlib import Path

from musicscore import Score, G, Chord, F
from unittest import TestCase
from musicxml import XMLGlissando, XMLSlide

types = [
    "normal glissando",
    "solid (+text)",
    "dashed",
    "dotted",
    "wavy",
    "normal slide",
    "solid (+text)",
    "dashed",
    "dotted",
    "wavy",
]


class TestLily33h(TestCase):
    def test_lily_33h_Spanners_Glissando(self):
        score = Score()
        part = score.add_part("p1")
        line_types = [None, "solid", "dashed", "dotted", "wavy"]
        for i, t in enumerate(types):
            chords = [Chord(G(4), 1), Chord(F(5), 1)]
            texts = t.split(" ")
            chords[0].add_lyric(texts[0])
            if len(texts) > 1:
                chords[1].add_lyric(texts[1])

            if i < 5:
                chords[0].add_x(XMLGlissando(type="start", line_type=line_types[i]))
                chords[1].add_x(XMLGlissando(type="stop"))
            else:
                chords[0].add_x(XMLSlide(type="start", line_type=line_types[i - 5]))
                chords[1].add_x(XMLSlide(type="stop"))
            [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from musicscore.accidental import *
from musicscore.beat import *
from musicscore.chord import *
from musicscore.clef import *
from musicscore.musictree import *
from musicscore.dynamics import *
from musicscore.key import *
from musicscore.layout import *
from musicscore.lyrics import *
from musicscore.measure import *
from musicscore.metronome import *
from musicscore.midi import *
from musicscore.note import *
from musicscore.part import *
from musicscore.quarterduration import *
from musicscore.score import *
from musicscore.staff import *
from musicscore.time import *
from musicscore.voice import *
from musicscore.simpleformat import *



import inspect

from musicscore import Chord, Score, beam_chord_group
from unittest import TestCase

from musicscore.tests.util import generate_path


class TestBeams(TestCase):
    def test_beams_example(self):
        score = Score()
        p = score.add_part("p1")
        p.add_measure([1, 4])
        for qd in 7 * 4 * [1 / 4]:
            p.add_chord(Chord(71, qd))
        for l in [
            [0, 71, 71, 71],
            [0, 71, 71, 0],
            [71, 0, 71, 71],
            [71, 71, 0, 71],
            [71, 0, 0, 71],
            [71, 0, 0, 71],
            [71, 0, 71, 0, 71],
            [71, 0, 71, 0, 71, 71],
            [71, 0, 0, 71, 71, 71, 71, 71],
        ]:
            for m in l:
                p.add_chord(Chord(m, 1 / len(l)))
        for i in range(1, len(p.get_children()) // 4):
            p.get_measure(i * 4 + 1).new_system = True
        m = p.get_measure(2)
        for chord in m.get_chords():
            chord.beams = None
        m = p.get_measure(3)
        chords = m.get_chords()
        chords[0].beams = {1: "begin", 2: "begin"}
        chords[1].beams = {1: "continue", 2: "continue"}
        chords[2].beams = {1: "end", 2: "end"}
        chords[3].beams = None
        m = p.get_measure(4)
        chords = m.get_chords()
        chords[0].beams = None
        chords[1].beams = {1: "begin", 2: "begin"}
        chords[2].beams = {1: "continue", 2: "continue"}
        chords[3].beams = {1: "end", 2: "end"}
        m = p.get_measure(5)
        beat = m.get_beats()[0]
        beat._update_chord_types()
        chords = beat.get_chords()
        beam_chord_group(chords[:2])
        beam_chord_group(chords[2:])

        m = p.get_measure(6)
        beat = m.get_beats()[0]
        beat._update_chord_types()
        chords = beat.get_chords()
        beam_chord_group(chords[:2])
        for chord in chords[2:]:
            chord.beams = None

        m = p.get_measure(7)
        beat = m.get_beats()[0]
        beat._update_chord_types()
        chords = beat.get_chords()
        beam_chord_group(chords[:2])
        beam_chord_group(chords[2:])
        chords[1].beams[1] = "continue"
        chords[2].beams[1] = "continue"

        m = p.get_measure(13)
        chords = m.get_chords()
        chords[0].beams = None
        chords[-1].beams = None

        path = generate_path(inspect.currentframe())
        score.export_xml(path)



from typing import Optional, Set

from musicscore.clef import Clef
from musicscore.exceptions import (
    StaffHasNoParentError,
    AlreadyFinalizedError,
    AddChordError,
)
from musicscore.finalize import FinalizeMixin
from musicscore.musictree import MusicTree
from musicscore.quantize import QuantizeMixin
from musicscore.voice import Voice
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLStaff

__all__ = ["Staff"]


class Staff(MusicTree, QuantizeMixin, FinalizeMixin, XMLWrapper):
    """
    Parent type: :obj:`~musicscore.measure.Measure`

    Child type: :obj:`~musicscore.voice.Voice`
    """

    _ATTRIBUTES = {"clef", "default_clef", "number"}
    _ATTRIBUTES = _ATTRIBUTES.union(MusicTree._ATTRIBUTES)
    _ATTRIBUTES = _ATTRIBUTES.union(QuantizeMixin._ATTRIBUTES)
    XMLClass = XMLStaff

    def __init__(self, number=None, clef=None, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(value_=1, **kwargs)
        self._number = None
        self._clef = None
        self.clef = clef
        self.number = number

    @property
    def clef(self) -> Clef:
        """
        :type: :obj:`~musicscore.clef.Clef`
        :return: Clef associated with the staff.
        :rtype: :obj:`~musicscore.clef.Clef`
        """
        if not self._clef:
            if self.get_previous_staff() and self.get_previous_staff().clef:
                self._clef = self.get_previous_staff().clef.__copy__()
                self._clef.show = False
        return self._clef

    @clef.setter
    def clef(self, val):
        if val is not None and not isinstance(val, Clef):
            raise TypeError
        if val and not val.number and self.clef:
            val.number = self.clef.number
        self._clef = val

    @property
    def number(self):
        """
        :type: None or int. If None number is set to 1.
        :return: positive int or None
        """
        if self._number is not None:
            return self.xml_object.value_
        else:
            return self._number

    @number.setter
    def number(self, val):
        self._number = val
        if val is None:
            self.xml_object.value_ = 1
        else:
            self.xml_object.value_ = val

    def add_child(self, child: Voice) -> Voice:
        """
        - Adds a :obj:`~musicscore.voice.Voice` as child to staff.
        - If voice number is ``None``, it is determined as length of children + 1.
        - If voice number is already set an is not equal to length of children + 1. a ``ValueError`` is raised.
        - If Staff has no parent :obj:`~musicscore.measure.Measure` a :obj:`~musicscore.exceptions.StaffHasNoParentError` is raised.
        - After adding voice its :obj:`~musicscore.voice.Voice.update_beats()` is called.

        :param child: :obj:`~musicscore.voice.Voice` , required
        :return: added :obj:`~musicscore.voice.Voice`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, "add_child")
        self._check_child_to_be_added(child)

        if not self.up:
            raise StaffHasNoParentError(
                "A child Voice can only be added to a Staff if staff has a Measure parent."
            )

        if child.number is not None and child.number != len(self.get_children()) + 1:
            raise ValueError(
                f"Voice number must be None or {len(self.get_children()) + 1}"
            )
        if child.number is None:
            child.number = len(self.get_children()) + 1
        else:
            child.number = len(self.get_children()) + 1

        child._parent = self
        self._children.append(child)

        return child

    def add_chord(self, *args, **kwargs):
        raise AddChordError()

    def add_voice(self, voice_number: Optional[int] = None) -> Voice:
        """
        - Creates and adds a new :obj:`~musicscore.voice.Voice` object as child to staff.
        - If voice number is greater than length of children + 1 all missing voices are created and added first.

        :param voice_number: positive int or None. If ``None`` voice number is determined as length of children + 1.
        :return: new :obj:`~musicscore.voice.Voice`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, "add_voice")
        if voice_number is None:
            voice_number = len(self.get_children()) + 1

        voice_object = self.get_voice(voice_number=voice_number)

        if voice_object is None:
            for _ in range(voice_number - len(self.get_children())):
                voice_object = self.add_child(Voice())
                voice_object.update_beats()
            return voice_object

        voice_object.update_beats()
        return voice_object

    def fill_with_rests(self):
        for voice in self.get_children():
            voice.fill_with_rests()

    def get_previous_staff(self) -> Optional["Staff"]:
        """
        :return: :obj:`Staff` with the same number in previous :obj:`~musicscore.measure.Measure` if existing else ``None``
        """
        if self.up and self.up.previous:
            my_index = self.up.get_children().index(self)
            try:
                return self.up.previous.get_children()[my_index]
            except IndexError:
                return None
        return None

    def get_last_pitch_steps_with_accidentals(self) -> Set[str]:
        """
        :return: A set of pitch steps with not natural accidental. This method is used to keep track of accidental signs which needs to
                 be shown hidden.
        :rtype:  Set[str]
        """
        output = set()
        for v in self.get_children():
            if v.get_chords():
                last_chord = v.get_chords()[-1]
                if not last_chord.is_rest:
                    for m in last_chord.midis:
                        if m.accidental.sign != "natural":
                            step = m.accidental.get_pitch_parameters()[0]
                            if step not in output:
                                output.add(step)
        return output



from pathlib import Path

from musicscore import Score, Time, Chord, B
from unittest import TestCase

"""
Compound time signatures of mixed type: (3+2)/8+3/4.
"""


class TestLily11e(TestCase):
    def test_lily_11e_TimeSignatures_CompoundMixed(self):
        score = Score()
        part = score.add_part("part")
        part.add_measure(time=Time("3+2", 8, 3, 4))
        [part.add_chord(Chord(B(4), x / 2)) for x in [1, 1, 1, 1, 1]]
        [part.add_chord(Chord(B(4), x)) for x in [1, 1, 1]]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
A 3/8 pickup measure, a measure that is split into one (incomplete, only 2/4) measure and an implicit measure,
and an incomplete measure (containg 3/4)
"""

from pathlib import Path

from musicscore import Score, Time, Chord, E, F, G, A, B, C, D
from unittest import TestCase


class TestLily46d(TestCase):
    def test_lily_46d_PicupMeasure_ImplicitMeasure(self):
        score = Score()
        part = score.add_part("p1")
        t = Time(4, 4)
        t.actual_signatures = [3, 8]
        part.add_measure(time=t)
        part.add_measure(Time(4, 4))
        for d in [2, 1]:
            part.add_chord(Chord(E(4), d * 1 / 2))
        for m in [F(4), G(4), A(4), B(4), C(5), D(5), 0]:
            part.add_chord(Chord(m, 1))

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Grace notes shall not mess up the lyrics, and they shall not be assigned a syllable.
"""

from pathlib import Path
from unittest import TestCase

from musicscore import Score, Chord, G, D, C, E, Lyrics
from musicscore.util import slur_chords


class TestLily61f(TestCase):
    def test_lily_61f_Lyrics_GraceNotes(self):
        score = Score()
        part = score.add_part("p1")
        chords = [
            Chord(G(4), 1),
            Chord(C(5), 1),
            Chord(C(5), 1),
            Chord(C(5), 1),
            Chord(C(5), 1),
            Chord(C(5), 1),
            Chord(C(5), 1),
            Chord(C(5), 1),
        ]
        chords[1].add_grace_chord(D(5))
        chords[3].add_grace_chord(D(5))
        chords[5].add_grace_chord(E(5))
        chords[5].add_grace_chord(D(5))
        chords[6].add_grace_chord(D(5))

        slur_chords([chords[0], chords[1]])
        slur_chords([chords[4], chords[5]])
        slur_chords([chords[1].get_grace_chords()[0], chords[1]], number=2)
        slur_chords([chords[3].get_grace_chords()[0], chords[3]], number=1)
        [part.add_chord(ch) for ch in chords]

        Lyrics([("Ly", None, None, "rics"), None, "on", ("Notes", None)]).add_to_chords(
            chords
        )

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



import inspect
from unittest import TestCase

from musicscore import C, B, G, Part, Chord, Score, Time
from musicscore.tests.util import generate_path
from musicxml.xmlelement.xmlelement import XMLNotehead


class TestMidiNotes(TestCase):
    def test_midi_note(self):
        m = C(4, "s")
        assert m.accidental.get_pitch_parameters() == ("C", 1, 4)
        m = C(0)
        assert m.accidental.get_pitch_parameters() == ("C", 0, 0)
        with self.assertRaises(ValueError):
            B(-1)
        with self.assertRaises(ValueError):
            G(9, "s")

    def test_midi_note_half_tone_accidentals(self):
        signs = [
            "double-flat",
            "flat-flat",
            "ff",
            "bb",
            "flat",
            "f",
            "b",
            "sharp",
            "s",
            "#",
            "double-sharp",
            "sharp-sharp",
            "ss",
            "##",
        ]
        for sign in signs:
            m = C(4, sign)
            if sign in ["double-flat", "flat-flat", "ff", "bb"]:
                assert m.value == 58
                assert m.accidental.get_pitch_parameters() == ("C", -2, 4)
                assert m.accidental.sign == "flat-flat"

            elif sign in ["flat", "f", "b"]:
                assert m.value == 59
                assert m.accidental.sign == "flat"
                assert m.accidental.get_pitch_parameters() == ("C", -1, 4)
            elif sign in ["sharp", "s", "#"]:
                assert m.value == 61
                assert m.accidental.sign == "sharp"
                assert m.accidental.get_pitch_parameters() == ("C", 1, 4)
            elif sign in ["double-sharp", "sharp-sharp", "ss", "##"]:
                assert m.value == 62
                assert m.accidental.sign == "double-sharp"
                assert m.accidental.get_pitch_parameters() == ("C", 2, 4)

    def test_midi_note_quarter_tones(self):
        m1 = C(4, "three-quarters-flat")
        assert m1.value == 60 - 3 / 2
        assert m1.accidental.sign == "three-quarters-flat"
        assert m1.accidental.get_pitch_parameters() == ("C", -1.5, 4)
        C(4, "quarter-flat")
        C(4, "quarter-sharp")
        C(4, "three-quarters-sharp")

    def test_midi_note_copy(self):
        m = C(4, "#")
        m.add_tie("start")
        copied = m.__copy__()
        assert m != copied
        assert m.value == copied.value
        assert id(m.accidental) == id(copied.accidental)
        assert id(m._ties) == id(copied._ties)

        copied = m.__deepcopy__()
        assert m != copied
        assert m.value == copied.value
        assert m.accidental != copied.accidental
        assert m.accidental.mode == copied.accidental.mode
        assert m.accidental.show == copied.accidental.show
        assert id(m._ties) != id(copied._ties)
        assert m._ties == copied._ties

        copied = m._copy_for_split()
        assert m != copied
        assert m.value == copied.value
        assert m.accidental != copied.accidental
        assert m.accidental.mode == copied.accidental.mode
        assert m.accidental.show == copied.accidental.show
        assert id(m._ties) != id(copied._ties)
        assert m._ties != copied._ties

        #
        # copied = m.__deepcopy__()
        # assert m != copied
        # assert m.value == copied.value
        # assert m.accidental != copied.accidental
        # assert m.accidental.mode == copied.accidental.mode
        # assert m.accidental.show == copied.accidental.show


class TestMidiNoteNoteHead(TestCase):
    def test_notehead_property(self):
        m = C(4)
        m.notehead = "square"
        assert isinstance(m.notehead, XMLNotehead)
        assert m.notehead.value_ == "square"

    def test_notehead_after_finalize(self):
        p = Part("p1")
        ch = Chord(C(4), 1)
        ch.midis[0].notehead = "square"
        p.add_chord(ch)
        p.finalize()
        assert ch.midis[0].parent_note.xml_notehead.value_ == "square"

    def test_notehead_copy_for_split(self):
        midi = C(4)
        midi.notehead = "square"
        copied = midi._copy_for_split()
        assert copied.notehead.value_ == "square"

    def test_midi_note_notehead_after_split(self):
        s = Score()
        p = s.add_part("p1")
        ch = Chord(C(4), 3)
        ch.midis[0].notehead = "square"
        p.add_measure(Time(2, 4))
        p.add_chord(ch)
        path = generate_path(inspect.currentframe())
        s.export_xml(path)
        assert p.get_chords()[1].midis[0].notehead.value_ == "square"
        assert p.get_chords()[1].midis[0].parent_note.xml_notehead.value_ == "square"



"""
Repeats can also be nested.
"""

from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase


class TestLily45c(TestCase):
    def test_lily_45d_RepeatsMultipleTimes(self):
        score = Score()
        p = score.add_part("p1")
        [p.add_chord(Chord(0, 4)) for _ in range(8)]

        score.set_multi_measure_rest(2, 3)
        score.set_multi_measure_rest(4, 7)

        p.get_measure(2).set_repeat_barline(location="left")
        p.get_measure(3).set_repeat_barline(location="right")

        p.get_measure(4).set_repeat_barline(location="left")
        p.get_measure(7).set_repeat_barline(location="right")

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from unittest import TestCase
from musicscore import Part, QuarterDuration
from musicscore.chord import GraceChord
from musicscore.exceptions import ChordException


class TestGraceChord(TestCase):
    def test_grace_chord_init(self):
        gch = GraceChord(60)
        assert gch.quarter_duration == 0
        assert isinstance(gch.quarter_duration, QuarterDuration)
        gch.quarter_duration = 0
        with self.assertRaises(ChordException):
            gch.quarter_duration = 2

    def test_grace_chord_type(self):
        p = Part("p1")
        gch = GraceChord(60)

        gch.type = "16th"
        p.add_chord(gch)
        p.finalize()
        assert p.get_chords()[0].notes[0].xml_type.value_ == "16th"



from unittest import TestCase

from musicscore import Chord


class TestTypes(TestCase):
    def test_types(self):
        ch = Chord(60, 1)
        assert ch.quarter_duration.get_type() == "quarter"
        assert ch.quarter_duration.get_number_of_dots() == 0
        assert ch.quarter_duration.get_tuplet_ratio() is None

        ch = Chord(60, 1 / 2)
        assert ch.quarter_duration.get_type() == "eighth"
        assert ch.quarter_duration.get_number_of_dots() == 0
        assert ch.quarter_duration.get_tuplet_ratio() is None

        ch = Chord(60, 1 / 2)
        for sd in [6, 12]:
            ch.quarter_duration.beat_subdivision = sd
            assert ch.quarter_duration.get_type() == "eighth"
            assert ch.quarter_duration.get_number_of_dots() == 1
            if sd == 6:
                assert ch.quarter_duration.get_tuplet_ratio() == (6, 4)
            else:
                assert ch.quarter_duration.get_tuplet_ratio() == (12, 8)

        ch = Chord(60, 1 / 3)
        assert ch.quarter_duration.get_type() == "eighth"
        assert ch.quarter_duration.get_number_of_dots() == 0
        assert ch.quarter_duration.get_tuplet_ratio() == (3, 2)

        ch = Chord(60, 2 / 3)
        ch.quarter_duration.beat_subdivision = 9
        assert ch.quarter_duration.get_type() == "eighth"
        assert ch.quarter_duration.get_number_of_dots() == 1
        assert ch.quarter_duration.get_tuplet_ratio() == (9, 8)



from unittest import TestCase
from musicscore import Score


class TestCore(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.parts = [self.score.add_part(x) for x in ["p1", "p2", "p3"]]
        self.measures = [self.parts[0].add_measure() for _ in range(3)]
        self.staves = [self.measures[0].get_children()[0]] + [
            self.measures[0].add_staff() for _ in range(3)
        ]
        self.voices = [self.staves[0].get_children()[0]] + [
            self.staves[0].add_voice() for _ in range(3)
        ]
        self.beats = self.voices[0].get_children()

    def test_score_get_descendents(self):
        for i in range(3):
            assert self.score.get_part(i + 1) == self.parts[i]
            assert self.score.get_part(part_number=i + 1) == self.parts[i]

        for i in range(3):
            assert self.score.get_measure(1, i + 1) == self.measures[i]
            assert (
                self.score.get_measure(part_number=1, measure_number=i + 1)
                == self.measures[i]
            )

        for i in range(4):
            assert self.score.get_staff(1, 1, i + 1) == self.staves[i]
            assert (
                self.score.get_staff(
                    part_number=1, measure_number=1, staff_number=i + 1
                )
                == self.staves[i]
            )

        for i in range(4):
            assert self.score.get_voice(1, 1, 1, i + 1) == self.voices[i]
            assert (
                self.score.get_voice(
                    part_number=1, measure_number=1, staff_number=1, voice_number=i + 1
                )
                == self.voices[i]
            )

        for i in range(4):
            assert self.score.get_beat(1, 1, 1, 1, i + 1) == self.beats[i]
            assert (
                self.score.get_beat(
                    part_number=1,
                    measure_number=1,
                    staff_number=1,
                    voice_number=1,
                    beat_number=i + 1,
                )
                == self.beats[i]
            )

    def test_part_get_descendents(self):
        first_part = self.parts[0]
        for i in range(3):
            assert first_part.get_measure(i + 1) == self.measures[i]
            assert first_part.get_measure(measure_number=i + 1) == self.measures[i]

        for i in range(4):
            assert first_part.get_staff(1, i + 1) == self.staves[i]
            assert (
                first_part.get_staff(measure_number=1, staff_number=i + 1)
                == self.staves[i]
            )

        for i in range(4):
            assert first_part.get_voice(1, 1, i + 1) == self.voices[i]
            assert (
                first_part.get_voice(
                    measure_number=1, staff_number=1, voice_number=i + 1
                )
                == self.voices[i]
            )

        for i in range(4):
            assert first_part.get_beat(1, 1, 1, i + 1) == self.beats[i]
            assert (
                first_part.get_beat(
                    measure_number=1, staff_number=1, voice_number=1, beat_number=i + 1
                )
                == self.beats[i]
            )

    def test_measure_get_descendents(self):
        first_measure = self.measures[0]

        for i in range(4):
            assert first_measure.get_staff(i + 1) == self.staves[i]
            assert first_measure.get_staff(staff_number=i + 1) == self.staves[i]

        for i in range(4):
            assert first_measure.get_voice(1, i + 1) == self.voices[i]
            assert (
                first_measure.get_voice(staff_number=1, voice_number=i + 1)
                == self.voices[i]
            )

        for i in range(4):
            assert first_measure.get_beat(1, 1, i + 1) == self.beats[i]
            assert (
                first_measure.get_beat(
                    staff_number=1, voice_number=1, beat_number=i + 1
                )
                == self.beats[i]
            )

    def test_staff_get_descendents(self):
        first_staff = self.staves[0]

        for i in range(4):
            assert first_staff.get_voice(i + 1) == self.voices[i]
            assert first_staff.get_voice(voice_number=i + 1) == self.voices[i]

        for i in range(4):
            assert first_staff.get_beat(1, i + 1) == self.beats[i]
            assert (
                first_staff.get_beat(voice_number=1, beat_number=i + 1) == self.beats[i]
            )

    def test_voice_get_descendents(self):
        first_voice = self.voices[0]
        for i in range(4):
            assert first_voice.get_beat(i + 1) == self.beats[i]
            assert first_voice.get_beat(beat_number=i + 1) == self.beats[i]

    def test_wrong_number_of_arguments_or_keys(self):
        with self.assertRaises(ValueError):
            self.score.get_beat(1)
        with self.assertRaises(ValueError):
            self.parts[0].get_beat(1)
        with self.assertRaises(ValueError):
            self.measures[0].get_beat(voice_number=2, beat_number=1)
        with self.assertRaises(ValueError):
            self.measures[0].get_beat(measure_number=1, beat_number=2)

    def test_wrong_value(self):
        with self.assertRaises(TypeError):
            self.measures[0].get_staff(0)
        with self.assertRaises(TypeError):
            self.measures[0].get_staff(-1)
        with self.assertRaises(TypeError):
            self.measures[0].get_voice(staff_number=1, voice_number=-2)



from typing import Optional

from musicxml.xmlelement.xmlelement import XMLKey

from musicscore.xmlwrapper import XMLWrapper

__all__ = ["Key"]


class Key(XMLWrapper):
    _ATTRIBUTES = {"fifths", "show"}
    XMLClass = XMLKey

    def __init__(self, fifths: int = 0, show: bool = True, *args, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(*args, **kwargs)
        self.fifths = fifths
        self._show = None
        self.show = show

    @property
    def fifths(self) -> Optional[int]:
        """
        Set and get ``value_`` of :obj:`~musicxml.xmlelement.xmlelement.XMLFifths` child object of associated :obj:`~musicxml.xmlelement.xmlelement.XMLKey`

        :return: ``self.xml_object.xml_fifths.value_``
        """
        if self.xml_object.xml_fifths:
            return self.xml_object.xml_fifths.value_

    @fifths.setter
    def fifths(self, val):
        self.xml_object.xml_fifths = val

    @property
    def show(self) -> bool:
        """
        Set and get show attribute. If ``False`` :obj:`Key` object is not shown.
        """
        return self._show

    @show.setter
    def show(self, val):
        if not isinstance(val, bool):
            raise TypeError
        self._show = val

    def __copy__(self):
        return self.__class__(fifths=self.fifths, show=self.show)



import math
from typing import Union, List

from musicscore.exceptions import (
    WrongNumberOfChordsError,
    LyricSyllabicOrExtensionError,
)
from musicxml.xmlelement.xmlelement import *

note_types = {
    (1, 12): "32nd",
    (1, 11): "32nd",
    (2, 11): "16th",
    (3, 11): "16th",
    (4, 11): "eighth",
    (6, 11): "eighth",
    (8, 11): "quarter",
    (1, 10): "32nd",
    (3, 10): "16th",
    (1, 9): "32nd",
    (2, 9): "16th",
    (4, 9): "eighth",
    (8, 9): "quarter",
    (1, 8): "32nd",
    (3, 8): "16th",
    (7, 8): "eighth",
    (1, 7): "16th",
    (2, 7): "eighth",
    (3, 7): "eighth",
    (4, 7): "quarter",
    (6, 7): "quarter",
    (1, 6): "16th",
    (1, 5): "16th",
    (2, 5): "eighth",
    (3, 5): "eighth",
    (4, 5): "quarter",
    (6, 5): "quarter",
    (8, 5): "half",
    (1, 4): "16th",
    (2, 4): "eighth",
    (3, 4): "eighth",
    (7, 4): "quarter",
    (4, 3): "half",
    (1, 3): "eighth",
    (2, 3): "quarter",
    (3, 2): "quarter",
    (1, 2): "eighth",
    (1, 1): "quarter",
    (2, 1): "half",
    (3, 1): "half",
    (4, 1): "whole",
    (6, 1): "whole",
    (8, 1): "breve",
    (12, 1): "breve",
}

#:
XML_ARTICULATION_CLASSES = [
    XMLAccent,
    XMLStrongAccent,
    XMLStaccato,
    XMLTenuto,
    XMLDetachedLegato,
    XMLStaccatissimo,
    XMLSpiccato,
    XMLScoop,
    XMLPlop,
    XMLDoit,
    XMLFalloff,
    XMLBreathMark,
    XMLCaesura,
    XMLStress,
    XMLUnstress,
]

#:
XML_TECHNICAL_CLASSES = [
    XMLUpBow,
    XMLDownBow,
    XMLHarmonic,
    XMLOpenString,
    XMLThumbPosition,
    XMLFingering,
    XMLPluck,
    XMLDoubleTongue,
    XMLTripleTongue,
    XMLStopped,
    XMLSnapPizzicato,
    XMLFret,
    XMLString,
    XMLHammerOn,
    XMLPullOff,
    XMLBend,
    XMLTap,
    XMLHeel,
    XMLToe,
    XMLFingernails,
    XMLHole,
    XMLArrow,
    XMLHandbell,
    XMLBrassBend,
    XMLFlip,
    XMLSmear,
    XMLOpen,
    XMLHalfMuted,
    XMLHarmonMute,
    XMLGolpe,
    XMLOtherTechnical,
]

#:
XML_ORNAMENT_CLASSES = [
    XMLDelayedInvertedTurn,
    XMLDelayedTurn,
    XMLHaydn,
    XMLInvertedMordent,
    XMLInvertedTurn,
    XMLInvertedVerticalTurn,
    XMLMordent,
    XMLOtherOrnament,
    XMLSchleifer,
    XMLShake,
    XMLTremolo,
    XMLTrillMark,
    XMLTurn,
    XMLVerticalTurn,
    XMLWavyLine,
]

#:
XML_DYNAMIC_CLASSES = [
    XMLF,
    XMLFf,
    XMLFff,
    XMLFfff,
    XMLFffff,
    XMLFfffff,
    XMLFp,
    XMLFz,
    XMLMf,
    XMLMp,
    XMLP,
    XMLPf,
    XMLPp,
    XMLPpp,
    XMLPppp,
    XMLPpppp,
    XMLPppppp,
    XMLRf,
    XMLRfz,
    XMLSf,
    XMLSffz,
    XMLSfp,
    XMLSfpp,
    XMLSfz,
    XMLSfzp,
    XMLOtherDynamics,
]

#:
XML_OTHER_NOTATIONS = [
    XMLArpeggiate,
    XMLFermata,
    XMLFootnote,
    XMLGlissando,
    XMLLevel,
    XMLNonArpeggiate,
    XMLOtherNotation,
    XMLSlide,
    XMLSlur,
]

#:
XML_DIRECTION_TYPE_CLASSES = [
    XMLRehearsal,
    XMLSegno,
    XMLCoda,
    XMLWords,
    XMLSymbol,
    XMLWedge,
    XMLDashes,
    XMLBracket,
    XMLPedal,
    XMLMetronome,
    XMLOctaveShift,
    XMLHarpPedals,
    XMLDamp,
    XMLDampAll,
    XMLEyeglasses,
    XMLStringMute,
    XMLScordatura,
    XMLPrincipalVoice,
    XMLPercussion,
    XMLAccordionRegistration,
    XMLStaffDivide,
    XMLOtherDirection,
]

#:
XML_ORNAMENT_AND_OTHER_NOTATIONS = [XMLAccidentalMark]

#:
XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS = [XMLDynamics]


def lcm(l):
    return math.lcm(*l)


def dToX(input_list, first_element=0):
    if isinstance(input_list, list) is False:
        raise TypeError("xToD(input_list)")
    else:
        output = [first_element]
        for i in range(len(input_list)):
            output.append(input_list[i] + output[i])
        return output


def xToD(input_list):
    result = []
    for i in range(1, len(input_list)):
        result.append(input_list[i] - input_list[i - 1])
    return result


def isinstance_as_string(
    child: object, parent_class_names: Union[str, List[str]]
) -> bool:
    """
    This function can be used to check if some class names (parent_class_names) can be found in another class's __mro__.
    If parent classes cannot be imported due to recursive imports this can be used instead of isinstance function.
    :param object child:
    :param str/[str] parent_class_names:
    :return: bool
    """
    if isinstance(parent_class_names, str):
        parent_class_names = [parent_class_names]

    for parent_class_name in parent_class_names:
        if parent_class_name not in [cls.__name__ for cls in child.__class__.__mro__]:
            return False
    return True


def _chord_is_in_a_repetition(chord):
    my_index = chord.up.up.get_chords().index(chord)
    if my_index > 0 and not chord.is_tied_to_previous:
        all_previous_chords = [
            chord.up.up.get_chords()[my_index - i] for i in range(1, my_index + 1)
        ]
        if set([ch.is_tied_to_previous for ch in all_previous_chords]) == {True}:
            return False
        previous_chord = all_previous_chords[0]
        if not previous_chord.is_rest and chord.has_same_pitches(previous_chord):
            return True
    return False


def slur_chords(chords, number=1, **kwargs):
    if len(chords) < 2:
        raise WrongNumberOfChordsError("util.slur_chords needs at list two chords.")

    chords[0].add_x(XMLSlur(type="start", number=number, **kwargs))
    chords[-1].add_x(XMLSlur(type="stop", number=number))
    for ch in chords[1:-1]:
        ch.add_x(XMLSlur(type="continue", number=number))


def trill_chords(chords, number=1, placement="above", **kwargs):
    if len(chords) < 2:
        raise WrongNumberOfChordsError("util.trill_chords needs at list two chords.")
    chords[0].add_x(XMLTrillMark(placement=placement, **kwargs))
    chords[0].add_x(
        XMLWavyLine(type="start", number=number, **kwargs), placement=placement
    )
    chords[-1].add_x(
        XMLWavyLine(type="stop", number=number, **kwargs), placement=placement
    )
    for ch in chords[1:-1]:
        ch.add_x(
            XMLWavyLine(type="continue", number=number, **kwargs), placement=placement
        )


def wedge_chords(chords, wedge_type, number=1, placement="below", **kwargs):
    if len(chords) < 2:
        raise WrongNumberOfChordsError("util.wedge_chords needs at list two chords.")

    chords[0].add_x(
        XMLWedge(type=wedge_type, number=number, **kwargs), placement=placement
    )
    chords[-1].add_x(XMLWedge(type="stop", number=number), placement=placement)
    for ch in chords[1:-1]:
        ch.add_x(XMLWedge(type="continue", number=number), placement=placement)


def bracket_chords(
    chords,
    line_type="solid",
    start_line_end="down",
    end_line_end="down",
    placement="above",
    number=1,
):
    if len(chords) < 2:
        raise WrongNumberOfChordsError("util.bracket_chords needs at list two chords.")

    chords[0].add_x(
        XMLBracket(
            type="start", line_end=start_line_end, line_type=line_type, number=number
        ),
        placement=placement,
    )
    chords[-1].add_x(
        XMLBracket(
            type="stop", line_end=end_line_end, line_type=line_type, number=number
        ),
        placement=placement,
    )
    for ch in chords[1:-1]:
        ch.add_x(
            XMLBracket(
                type="continue", line_end="none", line_type=line_type, number=number
            ),
            placement=placement,
        )


def octave_chords(chords, type="down", size=8, number=1):
    try:
        len(chords)
    except TypeError:
        chords = [chords]
    if type == "down":
        placement = "above"
    else:
        placement = "below"

    chords[0].add_x(
        XMLOctaveShift(type=type, size=size, number=number), placement=placement
    )
    chords[-1].add_x(
        XMLOctaveShift(type="stop", size=size, number=number), placement=placement
    )
    for ch in chords[1:-1]:
        ch.add_x(
            XMLOctaveShift(type="continue", size=size, number=number),
            placement=placement,
        )


def _generate_lyrics(lyrics, number=1, show_number=False, mode="list", **kwargs):
    def _get_syllables_extensions_from_group(syllabic_group):
        if syllabic_group[0] is None:
            raise LyricSyllabicOrExtensionError(
                "Syllabic or extension cannot be added to the beginning of a syllabic group."
            )
        extensions_ = []
        syllables_ = list(syllabic_group)[:]
        while syllables_[-1] is None:
            extensions_.append(syllables_.pop())
        return [syllables_, extensions_]

    if isinstance(lyrics, str) or isinstance(lyrics, tuple):
        lyrics = [lyrics]

    output = []
    if mode == "list":
        for i in range(len(lyrics)):
            ll = lyrics[i]
            if isinstance(ll, str):
                xl = XMLLyric(number=str(number), **kwargs)
                xl.xml_text = ll
                xl.xml_syllabic = "single"
                output.append(xl)

            elif isinstance(ll, tuple) or isinstance(ll, list):
                syllables, extensions = _get_syllables_extensions_from_group(ll)
                if len(syllables) == 1:
                    xl = XMLLyric(number=str(number), **kwargs)
                    xl.xml_text = syllables[0]
                    xl.xml_syllabic = "single"
                    if len(extensions) > 0:
                        xl.xml_extend = XMLExtend(type="start")

                    output.append(xl)
                else:
                    for j in range(len(syllables)):
                        xl = XMLLyric(number=str(number), **kwargs)
                        l = syllables[j]
                        if l:
                            xl.xml_text = l
                            if j == 0:
                                xl.xml_syllabic = "begin"
                            elif j == len(syllables) - 1:
                                xl.xml_syllabic = "end"
                                if len(extensions) > 0:
                                    xl.xml_extend = XMLExtend(type="start")
                            else:
                                xl.xml_syllabic = "middle"
                        else:
                            xl = None
                        output.append(xl)

                for k in range(len(extensions)):
                    xl = XMLLyric(number=str(number), **kwargs)
                    if k == len(extensions) - 1:
                        xl.xml_extend = XMLExtend(type="stop")
                    else:
                        xl.xml_extend = XMLExtend(type="continue")
                    xl.xsd_check = False
                    output.append(xl)
            elif ll is None:
                output.append(None)
            else:
                raise NotImplementedError(ll)
    else:
        raise NotImplementedError
    if show_number:
        first = XMLLyric(**kwargs)
        for k, i in output[0].attributes.items():
            setattr(first, k, i)
        first.add_child(XMLSyllabic("single"))
        first.add_child(XMLText(f"{number}."))
        first.add_child(XMLElision(" "))

        for child in output[0].get_children():
            first.add_child(child)

        output[0] = first
    return output


def split_list(original_list, split_indices):
    if not split_indices:
        return [original_list]

    if split_indices[0] == 0:
        split_indices = split_indices[1:]

    return [
        original_list[i:j] for i, j in zip([0] + split_indices, split_indices + [None])
    ]



from fractions import Fraction
from math import ceil, floor
from unittest import TestCase

from musicscore.beat import Beat
from musicscore.exceptions import QuarterDurationIsNotWritable
from musicscore.quarterduration import QuarterDuration, _check_quarter_duration_value


class TestQuarterDuration(TestCase):
    def test_quarter_duration_init(self):
        qd = QuarterDuration(3, 2)
        assert (qd.numerator, qd.denominator) == (3, 2)
        assert qd == 1.5
        assert qd == QuarterDuration(3, 2)
        qd = QuarterDuration(1.2)
        assert qd.as_integer_ratio() == (6, 5)
        assert qd == QuarterDuration(6, 5)
        assert qd == 1.2

    def test_quarter_duration_operands(self):
        qd = QuarterDuration(2)
        assert qd == 2
        assert qd <= 2
        assert qd >= 2
        assert not qd == 3
        assert qd < 3
        assert qd != 3
        assert qd > 1
        qd = QuarterDuration(1, 6)
        assert qd == Fraction(1, 6)
        assert qd == 1 / 6
        assert qd <= 1 / 6
        assert qd >= 1 / 6
        assert QuarterDuration(1, 3) == 1 / 3
        assert QuarterDuration(6 / 5) == 6 / 5
        assert QuarterDuration(6, 5) == 1.2
        assert QuarterDuration(1.2) == QuarterDuration(6, 5)
        assert QuarterDuration(Fraction(6 / 5)) == 6 / 5
        assert [
            QuarterDuration(7, 20),
            [QuarterDuration(3, 2), QuarterDuration(1)],
            QuarterDuration(23, 20),
        ] == [0.35, [1.5, 1], 1.15]
        assert [
            QuarterDuration(0.35),
            [QuarterDuration(1.5), QuarterDuration(1)],
            QuarterDuration(1.15),
        ] == [0.35, [1.5, 1], 1.15]

    def test_add_etc(self):
        assert isinstance(1 / 6 + QuarterDuration(1, 6), QuarterDuration)
        assert QuarterDuration(3, 4) + 0.25 == 1
        assert 1 / 6 + QuarterDuration(1, 6) == QuarterDuration(1, 3)
        assert QuarterDuration(1, 6) + 1 / 6 == QuarterDuration(1, 3)

        assert 1 / 6 + QuarterDuration(1, 6) == 1 / 3
        assert QuarterDuration(1, 6) + 1 / 6 == 1 / 3
        assert QuarterDuration(1, 6) + QuarterDuration(1, 6) == QuarterDuration(1, 3)
        assert QuarterDuration(1, 6) + QuarterDuration(1, 6) == 1 / 3
        assert QuarterDuration(1, 6) * 2 == 1 / 3
        assert 2 * QuarterDuration(1, 6) == 1 / 3

    def test_rtruediv(self):
        assert 2 / (1 / 2) == 4
        assert 2 / QuarterDuration(1, 2) == 4

    def test_split_quarter_duration(self):
        """
        Test if _get_beatwise_sections can split quarter duration to writable sections.
        """
        assert QuarterDuration(4)._get_beatwise_sections(
            beats=[Beat(1), Beat(1), Beat(1), Beat(1)]
        ) == [[4], None]
        assert QuarterDuration(3.5)._get_beatwise_sections(
            beats=[Beat(1), Beat(1), Beat(1), Beat(1)]
        ) == [[3, 0.5], None]
        assert QuarterDuration(3.5)._get_beatwise_sections(
            offset=0.5, beats=[Beat(1), Beat(1), Beat(1), Beat(1)]
        ) == [[0.5, 3], None]
        assert QuarterDuration(4)._get_beatwise_sections(
            offset=0.5, beats=[Beat(1), Beat(1), Beat(1), Beat(1)]
        ) == [[0.5, 3], 0.5]
        assert QuarterDuration(4)._get_beatwise_sections(
            offset=0.25, beats=[Beat(1), Beat(1), Beat(1)]
        ) == [[0.75, 2], 1.25]
        assert QuarterDuration(4)._get_beatwise_sections(
            offset=0.5, beats=[Beat(1), Beat(1.5), Beat(1.5), Beat(0.5)]
        ) == [[0.5, 3, 0.5], None]
        assert QuarterDuration(3)._get_beatwise_sections(
            offset=0.5, beats=[Beat(1), Beat(1), Beat(0.5), Beat(0.5)]
        ) == [[0.5, 2], 0.5]
        assert QuarterDuration(4)._get_beatwise_sections(
            offset=0.15, beats=[Beat(0.5), Beat(1.5), Beat(1)]
        ) == [[0.35, 1.5, 1], 1.15]

    def test_copy_quarter_duration(self):
        qd = QuarterDuration(3, 4)
        copied = qd.__copy__()
        assert id(qd.value) != id(copied.value)
        assert qd.value == copied.value

    def test_check_quarter_duration(self):
        assert _check_quarter_duration_value(1)
        assert _check_quarter_duration_value(1.2)
        assert _check_quarter_duration_value(1 / 3)
        assert _check_quarter_duration_value(Fraction(1, 3))
        assert _check_quarter_duration_value(QuarterDuration(1, 3))

    def test_get_type_and_dots(self):
        assert QuarterDuration(1)._get_type_and_dots() == ("quarter", 0)
        assert QuarterDuration(1.5)._get_type_and_dots() == ("quarter", 1)
        assert QuarterDuration(1.75)._get_type_and_dots() == ("quarter", 2)

        assert QuarterDuration(0.75)._get_type_and_dots() == ("eighth", 1)
        assert QuarterDuration(2 + 2 / 2 + 2 / 4)._get_type_and_dots() == ("half", 2)

        assert QuarterDuration(1 / 3)._get_type_and_dots() == ("eighth", 0)
        assert QuarterDuration(2 / 3)._get_type_and_dots() == ("quarter", 0)

    def test_get_type(self):
        assert QuarterDuration(1).get_type() == "quarter"
        assert QuarterDuration(1.5).get_type() == "quarter"
        assert QuarterDuration(1.75).get_type() == "quarter"

        assert QuarterDuration(0.75).get_type() == "eighth"
        assert QuarterDuration(2 + 2 / 2 + 2 / 4).get_type() == "half"

        assert QuarterDuration(1 / 3).get_type() == "eighth"
        assert QuarterDuration(2 / 3).get_type() == "quarter"

    def test_get_dots(self):
        assert QuarterDuration(1).get_number_of_dots() == 0
        assert QuarterDuration(1.5).get_number_of_dots() == 1
        assert QuarterDuration(1.75).get_number_of_dots() == 2

        assert QuarterDuration(0.75).get_number_of_dots() == 1
        assert QuarterDuration(2 + 2 / 2 + 2 / 4).get_number_of_dots() == 2

    def test_is_not_writable(self):
        for value in [0.23, 5 / 2, 9 / 2, 5 / 4, 9 / 5]:
            with self.assertRaises(QuarterDurationIsNotWritable):
                QuarterDuration(value)._get_type_and_dots()
            with self.assertRaises(QuarterDurationIsNotWritable):
                QuarterDuration(value).get_type()
            with self.assertRaises(QuarterDurationIsNotWritable):
                QuarterDuration(value).get_number_of_dots()

    def test_set_value(self):
        qt = QuarterDuration(2)
        qt.value = 3
        self.assertEqual(qt.value, 3)


class TestMagics(TestCase):
    cl = QuarterDuration

    def setUp(self):
        self.main = self.cl(70)
        self.equal = self.cl(70)
        self.equal_float = 70.0
        self.larger = self.cl(80)
        self.larger_float = 80.0
        self.smaller = self.cl(60)
        self.smaller_float = 60.0

    def test_abs(self):
        assert abs(self.cl(-70)) == 70

    def test_ceil(self):
        assert ceil(self.cl(70.2)) == 71

    def test_floor(self):
        assert floor(self.cl(70.2)) == 70

    def test_floor_division(self):
        a = self.cl(10)
        b = self.cl(4)
        c = self.cl(2)
        assert a // b == c
        assert a // 4 == c
        assert a // b == 2
        assert a // 4 == 2

    def test_gt(self):
        assert self.main > self.smaller
        assert self.main > self.smaller_float
        assert not self.main > self.equal
        assert not self.main > self.equal_float
        assert not self.main > self.larger
        assert not self.main > self.larger_float

    def test_ge(self):
        assert self.main >= self.smaller
        assert self.main >= self.smaller_float
        assert self.main >= self.equal
        assert self.main >= self.equal_float
        assert not self.main >= self.larger
        assert not self.main >= self.larger_float

    def test_le(self):
        assert not self.main <= self.smaller
        assert not self.main <= self.smaller_float
        assert self.main <= self.equal
        assert self.main <= self.equal_float
        assert self.main <= self.larger
        assert self.main <= self.larger_float

    def test_lt(self):
        assert not self.main < self.smaller
        assert not self.main < self.smaller_float
        assert not self.main < self.equal
        assert not self.main < self.equal_float
        assert self.main < self.larger
        assert self.main < self.larger_float

    def test_mod(self):
        a = self.cl(10)
        b = self.cl(3)
        c = self.cl(1)
        assert a % 3 == c
        assert a % 3 == 1
        assert a % b == c
        assert a % b == 1

    def test_mul(self):
        a = self.cl(10)
        b = self.cl(3)
        c = self.cl(30)
        assert a * b == 30
        assert a * b == c
        assert a * 3 == 30
        assert a * 3 == c

    def test_neg(self):
        a = self.cl(10)
        b = self.cl(-10)
        assert -a == b
        assert -a == -10
        assert -b == a
        assert -b == 10

    def test_pos(self):
        a = self.cl(10)
        assert +a == a

    def test_power(self):
        a = self.cl(10)
        b = self.cl(100)
        assert 10.0**2 == 100
        assert a**2 == 100
        assert a**2 == b

    def test_radd(self):
        a = self.cl(10)
        b = self.cl(100)
        assert a.__radd__(b) == b + a

    def test_rmod(self):
        a = self.cl(3)
        b = self.cl(10)
        assert a.__rmod__(b) == b % a

    def test_rmul(self):
        a = self.cl(3)
        b = self.cl(10)
        assert a.__rmul__(b) == b * a

    def test_eq(self):
        a = self.cl(10)
        b = self.cl(10)
        c = self.cl(11)
        assert a == b
        assert a == 10
        assert 10 == a
        assert a == 10.0
        assert a != 11
        assert a != c
        assert not a == c
        assert not a == 11
        assert not 11 == a
        d = self.cl(Fraction(10, 3))
        assert Fraction(10, 3) == Fraction(10, 3)
        assert Fraction(10, 3).__eq__(Fraction(10, 3))
        assert d.__eq__(Fraction(10, 3))
        assert d == Fraction(10, 3)
        assert a is not None
        assert a != None  # noqa

    def test_round(self):
        assert self.cl(70.7) == self.cl(70.7)
        assert round(self.cl(70.67), 1) == 70.7
        assert round(self.cl(70.67), 1) == self.cl(70.7)
        assert round(self.cl(70.67), 1) != self.cl(70.6)
        assert round(self.cl(70.67), 1) != 70.6

    def test_rtruediv(self):
        a = self.cl(3)
        b = self.cl(10)
        assert a.__rtruediv__(b) == Fraction(10, 3)

    def test_truediv(self):
        a = self.cl(10)
        b = self.cl(3)
        assert a / b == Fraction(10, 3)

    def test_trunc(self):
        a = self.cl(10.233)
        assert a.__trunc__() == 10

    def test_rfloordiv(self):
        a = self.cl(3)
        b = self.cl(10)
        assert a.__rfloordiv__(b) == b // a



from typing import List, Optional, Union, Tuple

from musicscore import Chord
from musicscore.exceptions import (
    VoiceIsFullError,
    AlreadyFinalizedError,
)
from musicscore.finalize import FinalizeMixin
from musicscore.measure import Measure
from musicscore.musictree import MusicTree
from musicscore.quantize import QuantizeMixin
from musicscore.time import Time
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLPart, XMLScorePart

__all__ = ["ScorePart", "Part"]


class ScorePart(XMLWrapper):
    _ATTRIBUTES = {"part"}

    XMLClass = XMLScorePart

    def __init__(self, part, *args, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._part = None
        self.part = part

    @property
    def part(self) -> "Part":
        """
        Setting part property updates its :obj:`~musicxml.xmlelement.xmlelement.XMLPartName` and sets or updates its :obj:`~musicscore.part.Id`

        :return: :obj:`~Part`
        """
        return self._part

    @part.setter
    def part(self, val):
        if not isinstance(val, Part):
            raise TypeError
        self._part = val
        self._update_id()
        self._update_name()
        self._update_abbreviation()

    def _update_abbreviation(self):
        self.xml_object.xml_part_abbreviation = self.part.abbreviation

    def _update_id(self):
        self.xml_object.id = self.part.id_

    def _update_name(self):
        self.xml_object.xml_part_name = self.part.name


class Part(MusicTree, QuantizeMixin, FinalizeMixin, XMLWrapper):
    """
    Parent type: :obj:`~musicscore.score.Score`

    Child type: :obj:`~musicscore.measure.Measure`
    """

    _ATTRIBUTES = {"id_", "name", "abbreviation"}
    _ATTRIBUTES = _ATTRIBUTES.union(MusicTree._ATTRIBUTES)
    _ATTRIBUTES = _ATTRIBUTES.union(QuantizeMixin._ATTRIBUTES)
    XMLClass = XMLPart

    def __init__(self, id, name=None, abbreviation=None, *args, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._id = None
        self.id_ = id
        self._name = None
        self.name = name
        self._abbreviation = None
        self.abbreviation = abbreviation
        self._score_part = ScorePart(part=self)
        self._current_measures = {}

    def _add_to_next_measure(self, current_measure, chord, staff_number, voice_number):
        if current_measure.next:
            current_measure = current_measure.next
        else:
            current_measure = self.add_measure()
        current_measure._add_chord(
            chord, staff_number=staff_number, voice_number=voice_number
        )
        return current_measure

    def _set_first_current_measure(self, staff_number, voice_number):
        for m in self.get_children():
            if m.get_voice(staff_number=staff_number, voice_number=voice_number):
                self.set_current_measure(staff_number, voice_number, m)
                return m

    @property
    def abbreviation(self) -> Optional[str]:
        return self._abbreviation

    @abbreviation.setter
    def abbreviation(self, val):
        self._abbreviation = val
        try:
            self.score_part._update_abbreviation()
        except AttributeError:
            pass

    @property
    def id_(self) -> str:
        return self.xml_object.id

    @id_.setter
    def id_(self, val):
        self.xml_object.id = val
        try:
            self.score_part._update_id()
        except AttributeError:
            pass

    @property
    def name(self) -> str:
        """
        Set and get name. Setting tries toupdate name of :obj:`musicscore.part.score_part`

        :type: str
        :return: part's name. If no name is set part's :obj:`id_` is returned.
        """
        if self._name is not None:
            return self._name
        else:
            return ""

    @name.setter
    def name(self, val):
        self._name = val
        try:
            self.score_part._update_name()
        except AttributeError:
            pass

    @property
    def score_part(self) -> ScorePart:
        """
        :return: the  :obj:`~musicscore.part.ScorePart` which is associated with this part.
        """
        return self._score_part

    def add_child(self, child: Measure) -> Measure:
        """
        Check and add child to list of children. Child's parent is set to self.

        :param child: :obj:`~musicscore.measure.Measure`
        :return: child
        :rtype: :obj:`~musicscore.measure.Measure`

        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, "add_child")
        super().add_child(child)
        self.xml_object.add_child(child.xml_object)
        return child

    def add_chord(
        self,
        chord: "Chord",
        *,
        staff_number: Optional[int] = None,
        voice_number: Optional[int] = 1,
    ) -> None:
        """
        - Adds a chord to the specified voice in current measure (see :obj:`get_current_measure()`).
        - If no current measure is set the first measure is selected.
        - If part has still no measures one measure is added.
        - If the specified voice in current measure is full chord is added to the voice with the same number in the next measure. If no
          next measure exists one measure is added
        - If a leftover chord remains after adding chord, it is added to voice's :obj:`~musicscore.voice.Voice.leftover_chord` and is
          added to so many next measures as needed.

        :param chord: :obj:`~musicscore.chord.Chord` required
        :param staff_number: positive int, None. If None is set to 1.
        :param voice_number: positive_int
        :return: None
        """
        if not isinstance(chord, Chord):
            raise TypeError(f"{chord} must be of type Chord.")
        if self._finalized is True:
            raise AlreadyFinalizedError(self, "add_chord")

        for gch in chord._grace_chords["before"]:
            self.add_chord(gch, staff_number=staff_number, voice_number=voice_number)

        if staff_number is None:
            staff_number = 1
        current_measure = self.get_current_measure(
            staff_number=staff_number, voice_number=voice_number
        )

        if not current_measure:
            if self.get_children():
                current_measure = self.get_children()[0]
            else:
                current_measure = self.add_measure()
        try:
            current_measure._add_chord(
                chord, staff_number=staff_number, voice_number=voice_number
            )
        except VoiceIsFullError:
            current_measure = self._add_to_next_measure(
                current_measure, chord, staff_number, voice_number
            )

        for gch in chord._grace_chords["after"]:
            self.add_chord(gch, staff_number=staff_number, voice_number=voice_number)

        leftover_chord = current_measure.get_voice(
            staff_number=staff_number, voice_number=voice_number
        ).leftover_chord
        while leftover_chord:
            current_measure = self._add_to_next_measure(
                current_measure, leftover_chord, staff_number, voice_number
            )
            leftover_chord = current_measure.get_voice(
                staff_number=staff_number, voice_number=voice_number
            ).leftover_chord

    def add_measure(
        self,
        time: Optional[Union[Time, List, Tuple]] = None,
        number: Optional[int] = None,
    ) -> Measure:
        """
        - Creates and adds a :obj:`~musicscore.measure.Measure` to part.
        - If time is not given last measure's :obj:`~musicscore.time.Time` is copied. Its :obj:`~musicscore.time.Time.show` property is set to
          ``False``
        - If number is not given last measure's number is incremented.
        - New measure's :obj:`~musicscore.key.Key` is a copy of last measure's :obj:`~musicscore.key.Key`. Its
          :obj:`~musicscore.key.Key.show` property is set to ``False``

        :param time: :obj:`~musicscore.time.Time`, (numerator, denominator), None
        :param number: positive int, None
        :return: created and added :obj:`~musicscore.measure.Measure`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, "add_measure")
        previous_measure = self.get_children()[-1] if self.get_children() else None
        if not time:
            if previous_measure:
                time = previous_measure.time.__copy__()
                time.show = False
            else:
                time = Time(4, 4)
        else:
            if not isinstance(time, Time):
                time = Time(*time)

        if not number:
            if previous_measure:
                number = previous_measure.number + 1
            else:
                number = 1

        m = Measure(number=number, time=time)
        child = self.add_child(m)
        if previous_measure:
            m.key = previous_measure.key.__copy__()
            m.key.show = False

            for staff in previous_measure.get_children():
                st = m.add_staff(staff_number=staff.number)
                if st.clef:
                    st.clef = staff.clef.__copy__()
                    st.clef.show = False
                if st:
                    st.add_voice(voice_number=1)
        else:
            m.add_voice(staff_number=None, voice_number=1)

        return child

    def get_current_measure(
        self, staff_number: Optional[int] = 1, voice_number: int = 1
    ):
        """
        Gets current measure for adding :obj:`~musicscore.chord.Chord` to a specific :obj:`~musicscore.voice.Voice`

        staff_number None is set to 1

        :param staff_number: positive int, None
        :param voice_number: positive int
        """
        if staff_number is None:
            staff_number = 1
        try:
            return self._current_measures[staff_number][voice_number]
        except KeyError:
            return self._set_first_current_measure(
                staff_number=staff_number, voice_number=voice_number
            )

    def set_current_measure(
        self, staff_number: int, voice_number: int, measure: Measure
    ) -> None:
        """
        Sets current measure for adding :obj:`~musicscore.chord.Chord` to a specific :obj:`~musicscore.voice.Voice`

        :param staff_number: positive int
        :param voice_number: positive int
        :param measure: :obj:`musicscore.measure.Measure`
        :return: None
        """
        if staff_number is None:
            staff_number = 1
        if not isinstance(measure, Measure):
            raise TypeError(f"{measure} must be of type 'Measure'.")
        if self._current_measures.get(staff_number):
            self._current_measures[staff_number][voice_number] = measure
        else:
            self._current_measures[staff_number] = {voice_number: measure}

    def finalize(self) -> None:
        if not self.get_children():
            self.add_measure()
        self.get_children()[-1].fill_with_rests()
        for beat in self.get_beats():
            if beat.get_quantized:
                beat.quantize_quarter_durations()
        super().finalize()



from pathlib import Path
from unittest import TestCase

from musicscore import Time
from musicscore.chord import Chord
from musicscore.score import Score
from musicscore.tests.util_subdivisions import (
    generate_all_quintuplets_manually,
    generate_all_triplets_manually,
)


class TestTuplets2(TestCase):
    def test_quarter_tuplets(self):
        score = Score()
        part = score.add_part("p1")
        t = Time(4, 4)
        t.actual_signatures = [1, 2, 1, 2]
        part.add_measure(t)
        tuplets = generate_all_triplets_manually() + generate_all_quintuplets_manually()

        tuplets = [x * 2 for t in tuplets for x in t]

        for t in tuplets:
            part.add_chord(Chord(60, t))

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Different Arpeggio directions (normal, up, down, non-arpeggiate)
"""

from pathlib import Path

from musicscore import Score, Chord, C, E, G
from unittest import TestCase


class TestLily32d(TestCase):
    def test_lily_32d_Arpeggio(self):
        score = Score()
        part = score.add_part("p1")

        types = ["normal", "up", "normal", "down", "normal", "none", "normal"]
        chords = [Chord([C(4), E(5), G(5)], 1) for _ in types]

        for t, ch in zip(types, chords):
            ch.arpeggio = t

        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)
